from jax.nn import (
    one_hot,
    softmax,
)
import jax.numpy as jnp

from fortuna.typing import Array


def focal_loss_fn(
    outputs: Array,
    targets: Array,
    gamma: float = 2.0,
) -> jnp.ndarray:
    """
    A focal loss function. See `[Mukhoti J. et a., 2020] <https://proceedings.neurips.cc/paper/2020/file/aeb7b30ef1d024a76f21a1d40e30c302-Paper.pdf>`_
    for reference.

    Parameters
    ----------
    outputs: Array
        Model outputs to be passed to the loss.
    targets: Array
        Target data points.
    gamma: float
        Hyper-parameter of the focal loss.

    Returns
    -------
    jnp.ndarray
        The focal loss evaluation.
    """
    probs = softmax(outputs, -1)
    targets = one_hot(targets, outputs.shape[-1])
    probs = jnp.sum(probs * targets, -1)
    return -jnp.mean((1 - probs) ** gamma * jnp.log(probs))



from functools import partial
import logging
from typing import (
    Dict,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
import flax.linen as nn
from flax.training.checkpoints import PyTree
import jax
from jax import random
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.model.model_manager.base import ModelManager
from fortuna.model.utils.random_features import RandomFeatureGaussianProcess
from fortuna.typing import (
    Array,
    Mutable,
    Params,
)
from fortuna.utils.data import get_inputs_from_shape
from fortuna.utils.nested_dicts import nested_update

logger = logging.getLogger(__name__)


class ClassificationModelManager(ModelManager):
    def __init__(self, model: nn.Module, model_editor: Optional[nn.Module] = None):
        r"""
        Classification model manager class. It orchestrates the forward pass of the model in the probabilistic model.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. The outputs must correspond to
            the logits of a softmax probability vector. The output dimension must be the same as the number of classes.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`f(w, x)`, where each component of :math:`f` corresponds to one of the classes.
        """
        super(ClassificationModelManager, self).__init__(model, model_editor)

    def apply(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        variables = params["model"].unfreeze()

        # setup dropout key
        if rng is not None:
            rng, dropout_rng = random.split(rng, 2)
            rngs = {"dropout": dropout_rng}
        else:
            rngs = None

        if mutable is not None:
            mutable_variables = mutable["model"].unfreeze()
            variables.update(mutable_variables)
            mutable = list(mutable_variables.keys())

        def apply_fn(v, x):
            _outputs = self.model.apply(
                v,
                x,
                mutable=mutable if mutable is not None else False,
                train=train,
                rngs=rngs,
            )
            if isinstance(_outputs, tuple) and not has_aux:
                _outputs = _outputs[0]
            return _outputs

        has_aux = train and mutable is not None

        if self.model_editor is not None:
            outputs = self.model_editor.apply(
                params["model_editor"],
                apply_fn=apply_fn,
                model_params=variables,
                x=inputs,
                has_aux=has_aux,
            )
        else:
            outputs = apply_fn(variables, inputs)

        if has_aux:
            outputs, mutable = outputs
            return outputs, {"mutable": FrozenDict({"model": mutable})}
        return outputs

    def init(
        self, input_shape: Tuple[int, ...], rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, Mapping]:
        if rng is None:
            rng = self.rng.get()
        rng, params_key, dropout_key = random.split(rng, 3)
        rngs = {"params": params_key, "dropout": dropout_key}
        params = dict(
            model=self.model.init(rngs, jnp.zeros((1,) + input_shape), **kwargs)
        )
        if self.model_editor is not None:
            if rng is None:
                rng = self.rng
            rng, params_key, dropout_key = random.split(rng, 3)
            rngs = {"params": params_key, "dropout": dropout_key}
            params.update(
                dict(
                    model_editor=self.model_editor.init(
                        rngs,
                        apply_fn=self.model.apply,
                        model_params=params["model"],
                        x=get_inputs_from_shape(input_shape),
                        has_aux=False,
                    )
                )
            )
        return params


class SNGPClassificationModelManagerMixin:
    def __init__(
        self,
        *args,
        output_dim: int,
        gp_hidden_features: int = 1024,
        normalize_input: bool = False,
        ridge_penalty: float = 1.0,
        momentum: Optional[float] = None,
        mean_field_factor: float = 1.0,
        **kwargs,
    ):
        """
        Classification model manager Mixin for SNGP models.

        Parameters
        ----------
        output_dim: int
            The output dimension of the network.
        normalize_input: bool
            Whether to normalize the input using nn.LayerNorm.
        gp_hidden_features: int
            The number of random fourier features.
        ridge_penalty: float
            Initial Ridge penalty to weight covariance matrix.
            This value is used to stabilize the eigenvalues of weight covariance estimate :math:`\Sigma` so that
            the matrix inverse can be computed for :math:`\Sigma = (\mathbf{I}*s+\mathbf{X}^T\mathbf{X})^{-1}`.
            The ridge factor :math:`s` cannot be too large since otherwise it will dominate
            making the covariance estimate not meaningful.
        momentum: Optional[float]
            A discount factor used to compute the moving average for posterior
            precision matrix. Analogous to the momentum factor in batch normalization.
            If `None` then update covariance matrix using a naive sum without
            momentum, which is desirable if the goal is to compute the exact
            covariance matrix by passing through data once (say in the final epoch).
            In this case, make sure to reset the precision matrix variable between
            epochs to avoid double counting.
        mean_field_factor: float
            The scale factor for mean-field approximation, used to adjust (at inference time) the influence of
            posterior variance in posterior mean approximation.
            See `Zhiyun L. et al., 2020 <https://arxiv.org/abs/2006.07584>`_ for more details.
        """
        super(SNGPClassificationModelManagerMixin, self).__init__(*args, **kwargs)
        self.output_dim = output_dim
        self.gp_hidden_features = gp_hidden_features
        self.normalize_input = normalize_input
        self.ridge_penalty = ridge_penalty
        self.momentum = momentum
        self.mean_field_factor = mean_field_factor
        self._gp_output_model = self._get_output_model()
        self._gp_output_model_mutable_keys = [
            "sngp_random_features",
            "sngp_laplace_covariance",
        ]

    def _get_output_model(self) -> RandomFeatureGaussianProcess:
        return RandomFeatureGaussianProcess(
            features=self.output_dim,
            hidden_features=self.gp_hidden_features,
            normalize_input=self.normalize_input,
            hidden_kwargs={"collection_name": "sngp_random_features"},
            covariance_kwargs={
                "ridge_penalty": self.ridge_penalty,
                "momentum": self.momentum,
                "collection_name": "sngp_laplace_covariance",
            },
        )

    def _mean_field_logits(
        self, logits: Array, covariance: Optional[Array] = None
    ) -> Array:
        """
        Adjust the model logits s.t. its softmax approximates the posterior mean
        (`Zhiyun L. et al., 2020 <https://arxiv.org/abs/2006.07584>`_).

        Parameters
        ----------
          logits: Array
            A float tensor of shape (batch_size, num_classes).
          covariance: Array
            A float tensor of shape (batch_size, batch_size) or (batch_size,). If None then it
            assumes the covariance is an identity matrix.

        Returns
        ----------
        Array
            The adjusted model logits.
        """
        # Implementation adapted from https://github.com/google/edward2/blob/520e28285e905e0021e49b52b982ee5ea170071c/edward2/tensorflow/layers/utils.py#L379
        if self.mean_field_factor < 0:
            raise ValueError(f"mean_field_factor cannot be < 0.")

        # Compute standard deviation.
        if covariance is None:
            variances = 1.0
        else:
            variances = jnp.diagonal(covariance) if covariance.ndim == 2 else covariance

        # Compute scaling coefficient for mean-field approximation.
        logits_scale = jnp.sqrt(1.0 + variances * self.mean_field_factor)

        # Cast logits_scale to compatible dimension.
        if len(logits.shape) > 1:
            logits_scale = jnp.expand_dims(logits_scale, axis=-1)

        return logits / logits_scale

    def apply(
        self,
        params: Params,
        inputs: Array,
        mutable: Optional[Mutable] = None,
        train: bool = False,
        rng: Optional[PRNGKeyArray] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, PyTree]]:
        if mutable and mutable is not None:
            deep_feature_extractor_mutable = {
                "model": FrozenDict(
                    {
                        k: v
                        for k, v in mutable["model"].items()
                        if k not in self._gp_output_model_mutable_keys
                    }
                )
            }
            gp_model_mutable = {
                k: v
                for k, v in mutable["model"].items()
                if k in self._gp_output_model_mutable_keys
            }
        else:
            deep_feature_extractor_mutable = mutable

        has_aux = mutable is not None and train

        def apply_fn(p, x):
            return super(SNGPClassificationModelManagerMixin, self).apply(
                p, x, deep_feature_extractor_mutable, train, rng
            )

        if self.model_editor is not None:
            deep_feature_extractor_outputs = self.model_editor.apply(
                params["model_editor"],
                apply_fn=apply_fn,
                model_params=params,
                x=inputs,
                has_aux=has_aux,
            )
        else:
            deep_feature_extractor_outputs = apply_fn(params, inputs)

        variables = params["model"].unfreeze()
        if mutable is not None:
            variables.update(gp_model_mutable)
            mutable = list(gp_model_mutable.keys())

        if has_aux:
            (
                deep_feature_extractor_outputs,
                deep_feature_extractor_mutable,
            ) = deep_feature_extractor_outputs

        if has_aux:
            outputs, gp_mutable = self._gp_output_model.apply(
                variables, deep_feature_extractor_outputs, mutable=mutable
            )
            outputs = outputs[0]
            if gp_mutable is not None:
                mutable = deep_feature_extractor_mutable["mutable"]["model"].unfreeze()
                mutable.update(gp_mutable.unfreeze())
                mutable = {"mutable": FrozenDict({"model": mutable})}
            return outputs, mutable
        else:
            logits, covariance = self._gp_output_model.apply(
                variables,
                deep_feature_extractor_outputs,
                mutable=False,
                return_full_covariance=False,
            )
            return self._mean_field_logits(logits, covariance)


class SNGPClassificationModelManager(
    SNGPClassificationModelManagerMixin, ClassificationModelManager
):
    def __init__(self, model: nn.Module, *args, **kwargs):
        """
        Classification model manager for SNGP models.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. The outputs of the model
            is the latent representation of the input, which in this case, does not correspond to the logits of a
            softmax probability vector. The output dimension of the model is not dependent on the number
            of classes in the classification task.
        """
        super(SNGPClassificationModelManager, self).__init__(model, *args, **kwargs)

    def init(
        self, input_shape: Tuple[int, ...], rng: Optional[PRNGKeyArray] = None, **kwargs
    ) -> Dict[str, FrozenDict]:
        if rng is None:
            rng = self.rng.get()
        rng, params_key, dropout_key = random.split(rng, 3)
        rngs = {"params": params_key, "dropout": dropout_key}
        model_params = self.model.init(rngs, jnp.zeros((1,) + input_shape), **kwargs)
        output_shape = jax.eval_shape(
            partial(self.model.apply, train=False),
            model_params,
            jnp.zeros((1,) + input_shape),
        ).shape
        if len(output_shape[1:]) > 1:  # drop batch size
            raise ValueError(
                f"The output shape for the given model is {output_shape}.\n"
                f"In order to use SNGP the output shape of the provide model has to be of shape"
                f"(batch_size, n_features)."
            )
        gp_params = self._gp_output_model.init(rngs, jnp.zeros(output_shape), **kwargs)
        params = dict(
            model=FrozenDict(
                nested_update(model_params.unfreeze(), gp_params.unfreeze())
            )
        )

        if self.model_editor is not None:
            rng, params_key, dropout_key = random.split(rng, 3)
            rngs = {"params": params_key, "dropout": dropout_key}
            params.update(
                dict(
                    model_editor=self.model_editor.init(
                        rngs,
                        apply_fn=self.model.apply,
                        model_params=params["model"],
                        x=get_inputs_from_shape(input_shape),
                        has_aux=False,
                    )
                )
            )
        return params



import jax.numpy as jnp

from fortuna.typing import Array


class MulticalibratorMixin:
    def mean_squared_error(self, values: Array, scores: Array) -> Array:
        """
        The mean squared error between the model evaluations and the scores.
        This is supposed to decrease at every round of the algorithm.

        Parameters
        ----------
        values: Array
            The model evaluations.
        scores: Array
            The scores.

        Returns
        -------
        Array
            The mean-squared error.
        """
        return self._loss_fn(values, scores)

    @staticmethod
    def _loss_fn(values: Array, scores: Array) -> Array:
        if scores.ndim == 2 and values.ndim == 1:
            scores = scores[:, 0]
            return jnp.mean(jnp.sum((values - scores) ** 2, axis=-1))
        return jnp.mean((values - scores) ** 2)



from typing import (
    Dict,
    Optional,
    Tuple,
    Union,
)

from jax import random
import jax.numpy as jnp

from fortuna.conformal.multivalid.iterative.multicalibrator import Multicalibrator
from fortuna.conformal.multivalid.mixins.classification.binary_multicalibrator import (
    BinaryClassificationMulticalibratorMixin,
)
from fortuna.typing import Array


class BinaryClassificationMulticalibrator(
    BinaryClassificationMulticalibratorMixin, Multicalibrator
):
    def calibrate(
        self,
        targets: Array,
        groups: Optional[Array] = None,
        probs: Optional[Array] = None,
        test_groups: Optional[Array] = None,
        test_probs: Optional[Array] = None,
        atol: float = 1e-4,
        rtol: float = 1e-6,
        min_prob_b: Union[float, str] = "auto",
        n_buckets: int = 100,
        n_rounds: int = 1000,
        eta: float = 0.1,
        split: float = 0.8,
        bucket_types: Tuple[str, ...] = ("<=", ">="),
        **kwargs,
    ) -> Union[Dict, Tuple[Array, Dict]]:
        return super().calibrate(
            scores=targets,
            groups=groups,
            values=probs,
            test_groups=test_groups,
            test_values=test_probs,
            atol=atol,
            rtol=rtol,
            min_prob_b=min_prob_b,
            n_buckets=n_buckets,
            n_rounds=n_rounds,
            eta=eta,
            split=split,
            bucket_types=bucket_types,
            **kwargs,
        )

    def apply_patches(
        self,
        groups: Optional[Array] = None,
        probs: Optional[Array] = None,
    ) -> Array:
        return super().apply_patches(groups=groups, values=probs)

    def calibration_error(
        self,
        targets: Array,
        groups: Optional[Array] = None,
        probs: Optional[Array] = None,
        n_buckets: int = 10000,
        **kwargs,
    ) -> Array:
        return super().calibration_error(
            scores=targets,
            groups=groups,
            values=probs,
        )

    def init_probs(self, size: int) -> Array:
        """
        Initialize probabilities.

        Parameters
        ----------
        size: int
            Number of data points.

        Returns
        -------
        Array
            A probability for each data point.
        """
        return self._maybe_init_values(values=None, size=size)

    def _maybe_init_values(self, values: Optional[Array], size: Optional[int] = None):
        if values is None:
            if size is None:
                raise ValueError(
                    "If `values` is not provided, `size` must be provided."
                )
            values = 0.5 * jnp.ones(size)
            values += 0.01 * random.normal(
                random.PRNGKey(self._seed), shape=values.shape
            )

        return jnp.copy(values)



import enum

from fortuna.model.model_manager.classification import (
    ClassificationModelManager,
    SNGPClassificationModelManager,
)
from fortuna.prob_model.posterior.deep_ensemble import DEEP_ENSEMBLE_NAME
from fortuna.prob_model.posterior.laplace import LAPLACE_NAME
from fortuna.prob_model.posterior.map import MAP_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld import CYCLICAL_SGLD_NAME
from fortuna.prob_model.posterior.sgmcmc.sghmc import SGHMC_NAME
from fortuna.prob_model.posterior.sngp import SNGP_NAME
from fortuna.prob_model.posterior.swag import SWAG_NAME


class ClassificationModelManagers(enum.Enum):
    """Map approximator name to model manager classes"""

    vars()[MAP_NAME] = ClassificationModelManager
    vars()[ADVI_NAME] = ClassificationModelManager
    vars()[DEEP_ENSEMBLE_NAME] = ClassificationModelManager
    vars()[LAPLACE_NAME] = ClassificationModelManager
    vars()[SWAG_NAME] = ClassificationModelManager
    vars()[SNGP_NAME] = SNGPClassificationModelManager
    vars()[SGHMC_NAME] = ClassificationModelManager
    vars()[CYCLICAL_SGLD_NAME] = ClassificationModelManager



# Code adapted from https://github.com/huggingface/transformers/blob/main/examples/flax/

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import flax
from jax import numpy as jnp
import numpy as np
from transformers import (
    PreTrainedTokenizerBase,
    TensorType,
)
from transformers.utils import PaddingStrategy


@flax.struct.dataclass
class FlaxDataCollatorForMultipleChoice:
    """
    Data collator used for multiple choice tasks like SWAG.
    Inputs are dynamically padded to either the maximum length of a batch or the provide `max_length`.

    Attributes
    ----------
    tokenizer: Union[:class:`~transformers.PreTrainedTokenizer`, :class:`~transformers.PreTrainedTokenizerFast`]
        The tokenizer used for encoding the data.
    padding: Union[bool, str, `~utils.PaddingStrategy`]:
        Select a strategy to pad the returned sequences (according to the model's padding side and padding
        index) among:

        - `True` or `'longest'`: Pad to the longest sequence in the batch (or no padding if only a single
        sequence if provided).
        - `'max_length'`: Pad to a maximum length specified with the argument `max_length` or to the maximum
        acceptable input length for the model if that argument is not provided.
        - `False` or `'do_not_pad'`: No padding (i.e., can output a batch with sequences of different
        lengths).
    max_length: Optional[int]
        Maximum length of the returned list and optionally padding length (see above).
    pad_to_multiple_of: Optional[int]
        If set will pad the sequence to a multiple of the provided value.

    """

    tokenizer: PreTrainedTokenizerBase
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None

    def __call__(self, features):
        label_name = "label" if "label" in features[0].keys() else "labels"
        labels = [feature.pop(label_name) for feature in features]
        batch_size = len(features)
        num_choices = len(features[0]["input_ids"])
        flattened_features = [
            [{k: v[i] for k, v in feature.items()} for i in range(num_choices)]
            for feature in features
        ]
        flattened_features = sum(flattened_features, [])

        batch = self.tokenizer.pad(
            flattened_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="np",
        )

        batch = {
            k: jnp.reshape(v, (batch_size, num_choices, -1)) for k, v in batch.items()
        }
        batch["labels"] = jnp.array(labels, dtype=jnp.int32)
        return batch


@flax.struct.dataclass
class FlaxDataCollatorForLanguageModeling:
    # Adapted from https://github.com/huggingface/transformers/blob/main/examples/flax/language-modeling/run_mlm_flax.py
    # and https://github.com/huggingface/transformers/blob/91b62f5a785264b626ec7ba708c894ac0276f5cb/src/transformers/data/data_collator.py#607
    """
    Data collator used for language modeling. Inputs are dynamically padded to the maximum length of a batch if they
    are not all of the same length.
    Attributes
    ----------
    tokenizer: Union[:class:`~transformers.PreTrainedTokenizer`, :class:`~transformers.PreTrainedTokenizerFast`]
        The tokenizer used for encoding the data.
    mlm: bool
        Whether or not to use masked language modeling. If set to `False`, the labels are the same as the inputs
        with the padding tokens ignored (by setting them to -100). Otherwise, the labels are -100 for non-masked
        tokens and the value to predict for the masked token. Default: True.
    mlm_probability: float
        The probability with which to (randomly) mask tokens in the input.
    .. note::
        For best performance, this data collator should be used with a dataset having items that are dictionaries or
        BatchEncoding, with the :obj:`"special_tokens_mask"` key, as returned by a
        :class:`~transformers.PreTrainedTokenizer` or a :class:`~transformers.PreTrainedTokenizerFast` with the
        argument :obj:`return_special_tokens_mask=True`.
    """

    tokenizer: PreTrainedTokenizerBase
    mlm: bool = True
    mlm_probability: float = 0.15

    def __post_init__(self):
        if self.mlm and self.tokenizer.mask_token is None:
            raise ValueError(
                "This tokenizer does not have a mask token which is necessary for masked language modeling. "
                "You should pass `mlm=False` to train on causal language modeling instead."
            )

    def __call__(
        self, examples: List[Dict[str, np.ndarray]], pad_to_multiple_of: int
    ) -> Dict[str, np.ndarray]:
        # Handle dict or lists with proper padding and conversion to tensor.
        batch = self.tokenizer.pad(
            examples,
            pad_to_multiple_of=pad_to_multiple_of,
            return_tensors=TensorType.NUMPY,
        )

        # If special token mask has been preprocessed, pop it from the dict.
        special_tokens_mask = batch.pop("special_tokens_mask", None)

        if self.mlm:
            batch["input_ids"], batch["labels"] = self.mask_tokens(
                batch["input_ids"], special_tokens_mask=special_tokens_mask
            )
        else:
            labels = batch["input_ids"]
            if self.tokenizer.pad_token_id is not None:
                # Replace self.tokenizer.pad_token_id with -100
                labels = np.where(labels == self.tokenizer.pad_token_id, -100, labels)
            else:
                labels = np.copy(labels)  # Makes a copy, just in case
            batch["labels"] = labels
        return batch

    def mask_tokens(
        self, inputs: np.ndarray, special_tokens_mask: Optional[np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Prepare masked tokens inputs/labels for masked language modeling: 80% MASK, 10% random, 10% original.
        """
        labels = inputs.copy()

        # We sample a few tokens in each sequence for MLM training (with probability `self.mlm_probability`)
        probability_matrix = np.full(labels.shape, self.mlm_probability)
        special_tokens_mask = special_tokens_mask.astype("bool")

        probability_matrix[special_tokens_mask] = 0.0
        masked_indices = np.random.binomial(1, probability_matrix).astype("bool")
        labels[~masked_indices] = -100  # We only compute loss on masked tokens

        # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
        indices_replaced = (
            np.random.binomial(1, np.full(labels.shape, 0.8)).astype("bool")
            & masked_indices
        )
        inputs[indices_replaced] = self.tokenizer.convert_tokens_to_ids(
            self.tokenizer.mask_token
        )

        # 10% of the time, we replace masked input tokens with random word
        indices_random = np.random.binomial(1, np.full(labels.shape, 0.5)).astype(
            "bool"
        )
        indices_random &= masked_indices & ~indices_replaced

        random_words = np.random.randint(
            self.tokenizer.vocab_size, size=labels.shape, dtype="i4"
        )
        inputs[indices_random] = random_words[indices_random]

        # The rest of the time (10% of the time) we keep the masked input tokens unchanged
        return inputs, labels



from typing import (
    Callable,
    Optional,
    Tuple,
)

import optax

from fortuna.typing import (
    AnyKey,
    Array,
    OptaxOptimizer,
)


class Optimizer:
    def __init__(
        self,
        method: Optional[OptaxOptimizer] = optax.adam(1e-2),
        n_epochs: int = 100,
        freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None,
    ):
        """
        An object to configure the optimization in the calibration process.

        Parameters
        ----------
        method: OptaxOptimizer
            An Optax optimizer.
        n_epochs: int
            Maximum number of epochs to run the calibration for.
        freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]]
            A callable taking in input a path in the nested dictionary of parameters, as well as the corresponding
            array of parameters, and returns "trainable" or "freeze", according to whether the corresponding parameter
            should be optimized or not.

            Examples
            --------
            .. code-block:: python
                def freeze_fun(path: Tuple[str], v: Array) -> str:
                    path = [p[:6] for p in path]  # take only the first 6 characters of each key"
                    return 'trainable' if "Dense" in path else 'frozen'`
        """
        self.method = method
        self.n_epochs = n_epochs
        self.freeze_fun = freeze_fun



import enum

from fortuna.prob_model.posterior.deep_ensemble import DEEP_ENSEMBLE_NAME
from fortuna.prob_model.posterior.deep_ensemble.deep_ensemble_posterior import (
    DeepEnsemblePosterior,
)
from fortuna.prob_model.posterior.laplace import LAPLACE_NAME
from fortuna.prob_model.posterior.laplace.laplace_posterior import LaplacePosterior
from fortuna.prob_model.posterior.map import MAP_NAME
from fortuna.prob_model.posterior.map.map_posterior import MAPPosterior
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_posterior import (
    ADVIPosterior,
)
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld import CYCLICAL_SGLD_NAME
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_posterior import (
    CyclicalSGLDPosterior,
)
from fortuna.prob_model.posterior.sgmcmc.sghmc import SGHMC_NAME
from fortuna.prob_model.posterior.sgmcmc.sghmc.sghmc_posterior import SGHMCPosterior
from fortuna.prob_model.posterior.sngp import SNGP_NAME
from fortuna.prob_model.posterior.sngp.sngp_posterior import SNGPPosterior
from fortuna.prob_model.posterior.swag import SWAG_NAME
from fortuna.prob_model.posterior.swag.swag_posterior import SWAGPosterior


class PosteriorApproximations(enum.Enum):
    """Map approximator name to posterior posterior_approximation."""

    vars()[MAP_NAME] = MAPPosterior
    vars()[ADVI_NAME] = ADVIPosterior
    vars()[DEEP_ENSEMBLE_NAME] = DeepEnsemblePosterior
    vars()[LAPLACE_NAME] = LaplacePosterior
    vars()[SWAG_NAME] = SWAGPosterior
    vars()[SNGP_NAME] = SNGPPosterior
    vars()[SGHMC_NAME] = SGHMCPosterior
    vars()[CYCLICAL_SGLD_NAME] = CyclicalSGLDPosterior



from typing import (
    List,
    Optional,
)

from jax import vmap
import jax.numpy as jnp

from fortuna.conformal.classification.base import ConformalClassifier
from fortuna.typing import Array


class SimplePredictionConformalClassifier(ConformalClassifier):
    def score(
        self,
        val_probs: Array,
        val_targets: Array,
    ) -> jnp.ndarray:
        """
        Compute score function.

        Parameters
        ----------
        val_probs: Array
            A two-dimensional array of class probabilities for each validation data point.
        val_targets: Array
            A one-dimensional array of validation target variables.

        Returns
        -------
        jnp.ndarray
            The conformal scores.
        """
        if val_probs.ndim != 2:
            raise ValueError(
                """`val_probs` must be a two-dimensional array. The first dimension is over the validation
            inputs. The second is over the classes."""
            )

        @vmap
        def score_fn(prob, target):
            return 1 - prob[target]

        return score_fn(val_probs, val_targets)

    def quantile(
        self,
        val_probs: Array,
        val_targets: Array,
        error: float = 0.05,
        scores: Optional[Array] = None,
    ) -> Array:
        """
        Compute a quantile of the scores.

        Parameters
        ----------
        val_probs: Array
            A two-dimensional array of class probabilities for each validation data point.
        val_targets: Array
            A one-dimensional array of validation target variables.
        error: float
            Coverage error. This must be a scalar between 0 and 1, extremes included.
        scores: Optional[Array]
            The conformal scores. This should be the output of
            :meth:`~fortuna.conformal.classification.simple_prediction.SimplePredictionConformalClassifier.score`.

        Returns
        -------
        float
            The conformal quantiles.
        """
        if error < 0 or error > 1:
            raise ValueError("""`error` must be a scalar between 0 and 1.""")

        if scores is None:
            scores = self.score(val_probs, val_targets)
        n = scores.shape[0]
        return jnp.quantile(scores, jnp.ceil((n + 1) * (1 - error)) / n)

    def conformal_set(
        self,
        val_probs: Array,
        test_probs: Array,
        val_targets: Array,
        error: float = 0.05,
        quantile: Optional[float] = None,
    ) -> List[List[int]]:
        """
        Coverage set of each of the test inputs, at the desired coverage error.

        Parameters
        ----------
        val_probs: Array
            A two-dimensional array of class probabilities for each validation data point.
        test_probs: Array
            A two-dimensional array of class probabilities for each test data point.
        val_targets: Array
            A one-dimensional array of validation target variables.
        error: float
            The coverage error. This must be a scalar between 0 and 1, extremes included.
        quantile: Optional[float]
            Conformal quantiles. This should be the output of
            :meth:`~fortuna.conformal.classification.simple_prediction.SimplePredictionConformalClassifier.quantile`.

        Returns
        -------
        List[List[int, ...]]
            The coverage sets.
        """
        if test_probs.ndim != 2:
            raise ValueError(
                """`test_probs` must be a two-dimensional array. The first dimension is over the validation
            inputs. The second is over the classes."""
            )

        if quantile is None:
            quantile = self.quantile(val_probs, val_targets, error)
        return [jnp.where(prob > 1 - quantile)[0].tolist() for prob in test_probs]



from typing import (
    Callable,
    NamedTuple,
)

import jax
import jax.numpy as jnp
from optax._src.base import PyTree

from fortuna.typing import Params

PreconditionerState = NamedTuple


class Preconditioner(NamedTuple):
    """A sampler preconditioner class.

    Attributes
    ----------
        init: Callable
            The state initialization function.
        update_preconditioner: Callable
            The state update function that takes gradients as an input.
        multiply_by_m_sqrt: Callable
            The function that multiples its input by the square root of mass matrix :math:`\sqrt{M}`.
        multiply_by_m_inv: Callable
            The function that multiples its input by the mass matrix inverse :math:`M^{-1}`.
        multiply_by_m_sqrt_inv: Callable
            The function that multiples its input by the square root of mass matrix inverse.

    """

    init: Callable[[Params], PreconditionerState]
    update_preconditioner: Callable[[PyTree, PreconditionerState], PreconditionerState]
    multiply_by_m_sqrt: Callable[[PyTree, PreconditionerState], PyTree]
    multiply_by_m_inv: Callable[[PyTree, PreconditionerState], PyTree]
    multiply_by_m_sqrt_inv: Callable[[PyTree, PreconditionerState], PyTree]


class RMSPropPreconditionerState(PreconditionerState):
    grad_moment_estimates: Params


def rmsprop_preconditioner(running_average_factor: float = 0.99, eps: float = 1.0e-7):
    """Create an instance of the adaptive RMSProp preconditioner.

    Parameters
    ----------
        running_average_factor: float
            The decay factor for the squared gradients moving average.
        eps: float
            :math:`\epsilon` constant for numerical stability.

    Returns
    -------
        preconditioner: Preconditioner
            An instance of RMSProp preconditioner.
    """

    def init_fn(params):
        return RMSPropPreconditionerState(
            grad_moment_estimates=jax.tree_util.tree_map(jnp.zeros_like, params)
        )

    def update_preconditioner_fn(gradient, preconditioner_state):
        grad_moment_estimates = jax.tree_util.tree_map(
            lambda e, g: e * running_average_factor
            + g**2 * (1 - running_average_factor),
            preconditioner_state.grad_moment_estimates,
            gradient,
        )
        return RMSPropPreconditionerState(grad_moment_estimates=grad_moment_estimates)

    def multiply_by_m_inv_fn(vec, preconditioner_state):
        return jax.tree_util.tree_map(
            lambda e, v: v / (eps + jnp.sqrt(e)),
            preconditioner_state.grad_moment_estimates,
            vec,
        )

    def multiply_by_m_sqrt_fn(vec, preconditioner_state):
        return jax.tree_util.tree_map(
            lambda e, v: v * jnp.sqrt(eps + jnp.sqrt(e)),
            preconditioner_state.grad_moment_estimates,
            vec,
        )

    def multiply_by_m_sqrt_inv_fn(vec, preconditioner_state):
        return jax.tree_util.tree_map(
            lambda e, v: v / jnp.sqrt(eps + jnp.sqrt(e)),
            preconditioner_state.grad_moment_estimates,
            vec,
        )

    return Preconditioner(
        init=init_fn,
        update_preconditioner=update_preconditioner_fn,
        multiply_by_m_inv=multiply_by_m_inv_fn,
        multiply_by_m_sqrt=multiply_by_m_sqrt_fn,
        multiply_by_m_sqrt_inv=multiply_by_m_sqrt_inv_fn,
    )


class IdentityPreconditionerState(PreconditionerState):
    pass


def identity_preconditioner():
    """Create an instance of no-op identity preconditioner.

    Returns
    -------
    preconditioner: Preconditioner
        An instance of identity preconditioner.
    """

    def init_fn(_):
        return IdentityPreconditionerState()

    def update_preconditioner_fn(*args, **kwargs):
        return IdentityPreconditionerState()

    def multiply_by_m_inv_fn(vec, _):
        return vec

    def multiply_by_m_sqrt_fn(vec, _):
        return vec

    def multiply_by_m_sqrt_inv_fn(vec, _):
        return vec

    return Preconditioner(
        init=init_fn,
        update_preconditioner=update_preconditioner_fn,
        multiply_by_m_inv=multiply_by_m_inv_fn,
        multiply_by_m_sqrt=multiply_by_m_sqrt_fn,
        multiply_by_m_sqrt_inv=multiply_by_m_sqrt_inv_fn,
    )



from typing import Optional

import jax.numpy as jnp

from fortuna.conformal.multivalid.mixins.multicalibrator import MulticalibratorMixin
from fortuna.typing import Array


class BinaryClassificationMulticalibratorMixin(MulticalibratorMixin):
    def mean_squared_error(self, probs: Array, targets: Array) -> Array:
        return super().mean_squared_error(values=probs, scores=targets)

    @staticmethod
    def _maybe_check_values(
        values: Optional[Array], test_values: Optional[Array] = None
    ):
        if values is not None:
            if values.ndim != 1:
                raise ValueError(
                    "`probs` must be a 1-dimensional array representing the probability that the "
                    "target variable is 1."
                )
            if jnp.any(values < 0) or jnp.any(values > 1):
                raise ValueError("All elements in `values` must be within [0, 1].")
        if test_values is not None:
            if test_values.ndim != 1:
                raise ValueError(
                    "`test_probs` must be a 1-dimensional array representing the probability that the "
                    "target variable is 1."
                )
            if jnp.any(test_values < 0) or jnp.any(test_values > 1):
                raise ValueError("All elements in `test_values` must be within [0, 1].")

    @staticmethod
    def _check_scores(scores: Array):
        if scores.ndim != 1:
            raise ValueError("`targets` must be a 1-dimensional array of integers.")
        if set(jnp.unique(scores).tolist()) != {0, 1}:
            raise ValueError("All values in `targets` must be 0 or 1.")
        if scores.dtype not in ["int32", "int64"]:
            raise ValueError("All elements in `targets` must be integers")



import enum

from fortuna.output_calib_model.state import OutputCalibState
from fortuna.prob_model.posterior.laplace.laplace_state import LaplaceState
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_state import ADVIState
from fortuna.prob_model.posterior.sgmcmc.cyclical_sgld.cyclical_sgld_state import (
    CyclicalSGLDState,
)
from fortuna.prob_model.posterior.sgmcmc.sghmc.sghmc_state import SGHMCState
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.prob_model.posterior.swag.swag_state import SWAGState


class NameToPosteriorState(enum.Enum):
    vars()[OutputCalibState.__name__] = OutputCalibState
    vars()[PosteriorState.__name__] = PosteriorState
    vars()[MAPState.__name__] = MAPState
    vars()[ADVIState.__name__] = ADVIState
    vars()[LaplaceState.__name__] = LaplaceState
    vars()[SWAGState.__name__] = SWAGState
    vars()[SGHMCState.__name__] = SGHMCState
    vars()[CyclicalSGLDState.__name__] = CyclicalSGLDState



from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

import flax.linen as nn
import jax.numpy as jnp

from fortuna.model_editor.base import ModelEditor
from fortuna.typing import (
    AnyKey,
    Array,
    InputData,
    Params,
)
from fortuna.utils.probit import sequential_probit_scaling


class ProbitModelEditor(ModelEditor):
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None
    top_k: Optional[int] = None
    memory: Optional[int] = None
    n_final_tokens: Optional[int] = None
    init_log_var: float = -5.0
    stop_gradient: bool = False

    @nn.compact
    def __call__(
        self,
        apply_fn: Callable[
            [Params, InputData], Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]
        ],
        model_params: Params,
        x: Any,
        has_aux: bool,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
        log_var = self.param(
            "log_var", nn.initializers.constant(self.init_log_var), (1,)
        )
        outputs = sequential_probit_scaling(
            apply_fn,
            model_params,
            x,
            log_var=log_var,
            has_aux=has_aux,
            freeze_fun=self.freeze_fun,
            top_k=self.top_k,
            memory=self.memory,
            n_final_tokens=self.n_final_tokens,
            stop_gradient=self.stop_gradient,
        )
        return outputs



class CalibProcessor:
    def __init__(
        self,
        devices: int = -1,
        disable_jit: bool = False,
        n_posterior_samples: int = 30,
    ):
        """
        An object to configure computational aspects of the calibration process.

        Parameters
        ----------
        devices: int
            A list of devices to be used during training.
            At the moment two options are supported: use all devices (`devices=-1`) or use no device (`devices=0`).
        disable_jit: bool
            if True, no function within the calibration loop is jitted.
        n_posterior_samples: int
            Number of posterior samples to draw from the posterior distribution for the calibration process.
        """
        self.devices = devices
        self.disable_jit = disable_jit
        self.n_posterior_samples = n_posterior_samples



from __future__ import annotations

import logging
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from jax._src.prng import PRNGKeyArray
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
import numpy as np

from fortuna.data.loader import (
    DataLoader,
    InputsLoader,
)
from fortuna.distribution.gaussian import DiagGaussian
from fortuna.prob_model.fit_config.base import FitConfig
from fortuna.prob_model.joint.base import Joint
from fortuna.prob_model.joint.state import JointState
from fortuna.prob_model.posterior.base import Posterior
from fortuna.prob_model.posterior.map.map_state import MAPState
from fortuna.prob_model.posterior.normalizing_flow.advi import ADVI_NAME
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_approximator import (
    ADVIPosteriorApproximator,
)
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_architecture import (
    ADVIArchitecture,
)
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_state import ADVIState
from fortuna.prob_model.posterior.normalizing_flow.advi.advi_trainer import (
    ADVITrainer,
    JittedADVITrainer,
    MultiDeviceADVITrainer,
)
from fortuna.prob_model.posterior.posterior_state_repository import (
    PosteriorStateRepository,
)
from fortuna.prob_model.posterior.run_preliminary_map import run_preliminary_map
from fortuna.typing import (
    AnyKey,
    Array,
    OptaxOptimizer,
    Params,
    Status,
)
from fortuna.utils.builtins import get_dynamic_scale_instance_from_model_dtype
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.freeze import get_trainable_paths
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
    nested_unpair,
)
from fortuna.utils.strings import decode_encoded_tuple_of_lists_of_strings_to_array


class ADVIPosterior(Posterior):
    def __init__(
        self,
        joint: Joint,
        posterior_approximator: ADVIPosteriorApproximator,
    ):
        """
        Automatic Differentiation Variational Inference (ADVI) approximate posterior class.

        Parameters
        ----------
        joint: Joint
            A joint distribution object.
        posterior_approximator: ADVI
            An ADVI posterior approximator.
        """
        super().__init__(joint=joint, posterior_approximator=posterior_approximator)
        self._base = None
        self._architecture = None
        self._unravel = None
        self._indices = None

    def __str__(self):
        return ADVI_NAME

    def fit(
        self,
        train_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        fit_config: FitConfig = FitConfig(),
        map_fit_config: Optional[FitConfig] = None,
        **kwargs,
    ) -> Dict[str, Status]:
        super()._checks_on_fit_start(fit_config, map_fit_config)

        status = dict()

        if super()._is_state_available_somewhere(fit_config):
            state = super()._restore_state_from_somewhere(
                fit_config=fit_config, allowed_states=(MAPState, ADVIState)
            )

        elif super()._should_run_preliminary_map(fit_config, map_fit_config):
            state, status["map"] = run_preliminary_map(
                joint=self.joint,
                train_data_loader=train_data_loader,
                val_data_loader=val_data_loader,
                map_fit_config=map_fit_config,
                rng=self.rng,
                **kwargs,
            )
        else:
            state = None

        state, log_stds = self._init_map_state(
            state=state, data_loader=train_data_loader, fit_config=fit_config
        )

        if fit_config.optimizer.freeze_fun is not None:
            which_params = get_trainable_paths(
                params=state.params, freeze_fun=fit_config.optimizer.freeze_fun
            )
        else:
            which_params = None

        rav, self._unravel, sub_unravel, rav_log_stds = self._get_unravel(
            params=state.params, log_stds=log_stds, which_params=which_params
        )

        size_rav = len(rav)
        self._base, self._architecture = self._get_base_and_architecture(size_rav)

        trainer_cls = select_trainer_given_devices(
            devices=fit_config.processor.devices,
            base_trainer_cls=ADVITrainer,
            jitted_trainer_cls=JittedADVITrainer,
            multi_device_trainer_cls=MultiDeviceADVITrainer,
            disable_jit=fit_config.processor.disable_jit,
        )

        trainer = trainer_cls(
            predict_fn=self.joint.likelihood.prob_output_layer.predict,
            save_checkpoint_dir=fit_config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=fit_config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=fit_config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=fit_config.monitor.disable_training_metrics_computation,
            eval_every_n_epochs=fit_config.monitor.eval_every_n_epochs,
            early_stopping_monitor=fit_config.monitor.early_stopping_monitor,
            early_stopping_min_delta=fit_config.monitor.early_stopping_min_delta,
            early_stopping_patience=fit_config.monitor.early_stopping_patience,
            freeze_fun=fit_config.optimizer.freeze_fun,
            base=self._base,
            architecture=self._architecture,
            which_params=which_params,
            unravel=self._unravel,
            sub_unravel=sub_unravel,
        )

        state = self._init_advi_from_map_state(
            rav=rav,
            rav_log_stds=rav_log_stds,
            state=state,
            init_params=self._architecture.init_params,
            optimizer=fit_config.optimizer.method,
        )

        logging.info("Run ADVI.")
        state, status["advi"] = trainer.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=self.joint._batched_negative_log_joint_prob,
            training_dataloader=train_data_loader,
            training_dataset_size=train_data_loader.size,
            n_epochs=fit_config.optimizer.n_epochs,
            metrics=fit_config.monitor.metrics,
            validation_dataloader=val_data_loader,
            validation_dataset_size=val_data_loader.size
            if val_data_loader is not None
            else None,
            verbose=fit_config.monitor.verbose,
            unravel=self._unravel,
            n_samples=self.posterior_approximator.n_loss_samples,
            callbacks=fit_config.callbacks,
            max_grad_norm=fit_config.hyperparameters.max_grad_norm,
            gradient_accumulation_steps=fit_config.hyperparameters.gradient_accumulation_steps,
        )

        self.state = PosteriorStateRepository(
            fit_config.checkpointer.save_checkpoint_dir
            if fit_config.checkpointer.dump_state is True
            else None
        )
        self.state.put(state, keep=fit_config.checkpointer.keep_top_n_checkpoints)
        logging.info("Fit completed.")
        return status

    def _get_base_and_architecture(
        self, size_rav: int
    ) -> Tuple[DiagGaussian, ADVIArchitecture]:
        base = DiagGaussian(
            mean=jnp.zeros(size_rav),
            std=jnp.exp(self.posterior_approximator.log_std_base) * jnp.ones(size_rav),
        )
        architecture = ADVIArchitecture(
            size_rav, std_init_params=self.posterior_approximator.std_init_params
        )
        return base, architecture

    def sample(
        self,
        rng: Optional[PRNGKeyArray] = None,
        input_shape: Optional[Tuple[int, ...]] = None,
        inputs_loader: Optional[InputsLoader] = None,
        inputs: Optional[Array] = None,
        **kwargs,
    ) -> JointState:
        """
        Sample from the posterior distribution. Either `input_shape` or `_inputs_loader` must be passed.
        Parameters
        ----------
        rng : Optional[PRNGKeyArray]
            A random number generator. If not passed, this will be taken from the attributes of this class.
        input_shape: Optional[Tuple[int, ...]]
            Shape of a single input.
        inputs_loader: Optional[InputsLoader]
            Input data loader. If `input_shape` is passed, then `inputs` and `inputs_loader` are ignored.
        inputs: Optional[Array]
            Input variables.
        Returns
        -------
        JointState
            A sample from the posterior distribution.
        """
        if rng is None:
            rng = self.rng.get()
        state = self.state.get()

        if self._base is None or self._unravel is None:
            if state._encoded_which_params is None:
                n_params = len(ravel_pytree(state.params)[0]) // 2
                which_params = None
            else:
                which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
                    state._encoded_which_params
                )
                n_params = len(
                    ravel_pytree(
                        nested_unpair(
                            d=state.params.unfreeze(),
                            key_paths=which_params,
                            labels=("mean", "log_std"),
                        )[1]
                    )[0]
                )
            _base, _architecture = self._get_base_and_architecture(n_params)
            _unravel = self._get_unravel(
                FrozenDict(
                    nested_unpair(
                        d=state.params.unfreeze(),
                        key_paths=which_params,
                        labels=("mean", "log_std"),
                    )[0]
                    if which_params
                    else {
                        k: dict(params=v["params"]["mean"])
                        for k, v in state.params.items()
                    }
                ),
                which_params=which_params,
            )[1]

            self._base = _base
            self._architecture = _architecture
            self._unravel = _unravel
        else:
            _base = self._base
            _architecture = self._architecture
            _unravel = self._unravel

        if state._encoded_which_params is None:
            means = _unravel(
                _architecture.forward(
                    {
                        s: ravel_pytree(
                            {k: v["params"][s] for k, v in state.params.items()}
                        )[0]
                        for s in ["mean", "log_std"]
                    },
                    _base.sample(rng),
                )[0][0]
            )
        else:
            which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
                state._encoded_which_params
            )
            means, log_stds = nested_unpair(
                d=state.params.unfreeze(),
                key_paths=which_params,
                labels=("mean", "log_std"),
            )
            rav_params = {
                k: ravel_pytree([nested_get(d=d, keys=path) for path in which_params])[
                    0
                ]
                for k, d in zip(["mean", "log_std"], [means, log_stds])
            }
            rav_params = _architecture.forward(params=rav_params, u=_base.sample(rng))[
                0
            ][0]

            means = _unravel(rav_params)

        return JointState(
            params=FrozenDict(means),
            mutable=state.mutable,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
        )

    def _init_map_state(
        self,
        state: Optional[Union[MAPState, ADVIState]],
        data_loader: DataLoader,
        fit_config: FitConfig,
    ) -> Tuple[MAPState, Params]:
        if state is None:
            state = super()._init_joint_state(data_loader)

        log_stds = None

        if isinstance(state, ADVIState):
            if state._encoded_which_params is not None:
                which_params = decode_encoded_tuple_of_lists_of_strings_to_array(
                    state._encoded_which_params
                )
                means, log_stds = nested_unpair(
                    d=state.params.unfreeze(),
                    key_paths=which_params,
                    labels=("mean", "log_std"),
                )
                means, log_stds = FrozenDict(means), FrozenDict(log_stds)
            else:
                means, log_stds = [
                    FrozenDict(
                        {
                            k: dict(params=v["params"][s])
                            for k, v in state.params.items()
                        }
                    )
                    for s in ["mean", "log_std"]
                ]
            state = state.replace(params=means)
            del means

        state = MAPState.init(
            params=state.params,
            mutable=state.mutable,
            optimizer=fit_config.optimizer.method,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
            dynamic_scale=get_dynamic_scale_instance_from_model_dtype(
                getattr(self.joint.likelihood.model_manager.model, "dtype")
                if hasattr(self.joint.likelihood.model_manager.model, "dtype")
                else None
            ),
        )

        return state, log_stds

    def _init_advi_from_map_state(
        self,
        rav: Array,
        rav_log_stds: Optional[Array],
        state: MAPState,
        init_params: Callable,
        optimizer: OptaxOptimizer,
    ) -> ADVIState:
        return ADVIState.init(
            params=FrozenDict(
                init_params(self.rng.get(), mean=rav, log_std=rav_log_stds),
            ),
            mutable=state.mutable,
            optimizer=optimizer,
            calib_params=state.calib_params,
            calib_mutable=state.calib_mutable,
            dynamic_scale=get_dynamic_scale_instance_from_model_dtype(
                getattr(self.joint.likelihood.model_manager.model, "dtype")
                if hasattr(self.joint.likelihood.model_manager.model, "dtype")
                else None
            ),
        )

    def _get_unravel(
        self,
        params,
        log_stds: Optional[Params] = None,
        which_params: Optional[Tuple[List[AnyKey], ...]] = None,
    ):
        if which_params is None:
            rav, unravel = ravel_pytree(params)
            rav_log_stds = ravel_pytree(log_stds)[0] if log_stds is not None else None
            sub_unravel = None
        else:
            rav, sub_unravel = ravel_pytree(
                tuple([nested_get(params, path) for path in which_params])
            )

            def unravel(_rav):
                return FrozenDict(
                    nested_set(
                        d=params.unfreeze(),
                        key_paths=which_params,
                        objs=sub_unravel(_rav),
                    )
                )

            if log_stds is not None:
                rav_log_stds = ravel_pytree(
                    nested_set(
                        d={},
                        key_paths=which_params,
                        objs=tuple(
                            [nested_get(log_stds, path) for path in which_params]
                        ),
                        allow_nonexistent=True,
                    )
                )[0]
            else:
                rav_log_stds = None

        return rav, unravel, sub_unravel, rav_log_stds



import flax.linen as nn
import jax.numpy as jnp

from fortuna.typing import Array


class ScalarHyperparameterModel(nn.Module):
    r"""
    A scalar hyperparameter model. The scalar value of the hyperparameter will not change during training, and it will
    be broadcasted to the output dimension.

    Parameters
    ----------
    output_dim: int
        The output model dimension.
    value: float
        Scalar value of the hyperparameter.
    """
    output_dim: int
    value: float

    def setup(self) -> None:
        if type(self.value) != float:
            raise ValueError(
                f"`value` must be a float, but a {type(self.value)} was found instead."
            )
        dummy = self.param("none", nn.initializers.zeros, (0,))

    def __call__(self, x: Array, **kwargs) -> jnp.ndarray:
        return jnp.broadcast_to(self.value, shape=(x.shape[0], self.output_dim))



import pandas as pd
import numpy as np

data_types = ["val", "test"]
tp_threshold = 0.97
tn_threshold = 0.99

for data_type in data_types:
    print(f"Data type: {data_type}")
    df = pd.read_csv(f"~/Downloads/cpp_results_{data_type}.csv")
    dfmc = pd.read_csv(f"~/Downloads/cpp_results_{data_type}_with_mc.csv")
    df = df.dropna(axis=0).astype(float)
    dfmc = dfmc.dropna(axis=0).astype(float)

    print(f"Average change in tp_precision after new calibration: {(df['tp_precision.1'] - df['tp_precision']).mean()}")
    print(f"Average change in tp_precision after multicalibrate + new calibration: {(dfmc['tp_precision.1'] - dfmc['tp_precision']).mean()}\n")

    print(f"Average change in tn_precision after new calibration: {(df['tn_precision.1'] - df['tn_precision']).mean()}")
    print(f"Average change in tn_precision after multicalibrate + new calibration: {(dfmc['tn_precision.1'] - dfmc['tn_precision']).mean()}\n")

    print(f"Average change in tp_percent after new calibration: {(df['tp_percent.1'] - df['tp_percent']).mean()}")
    print(f"Average change in tp_percent after multicalibrate + new calibration: {(dfmc['tp_percent.1'] - dfmc['tp_percent']).mean()}\n")

    print(f"Average change in tn_percent after new calibration: {(df['tn_percent.1'] - df['tn_percent']).mean()}")
    print(f"Average change in tn_percent after multicalibrate + new calibration: {(dfmc['tn_percent.1'] - dfmc['tn_percent']).mean()}\n")

    print(f"Average change in total_coverage after new calibration: {(df['total_coverage.1'] - df['total_coverage']).mean()}")
    print(f"Average change in total_coverage after multicalibrate + new calibration: {(dfmc['total_coverage.1'] - dfmc['total_coverage']).mean()}\n")

    print(f"Average change in total_ar after new calibration: {(df['total_ar.1'] - df['total_ar']).mean()}")
    print(f"Average change in total_ar after multicalibrate + new calibration: {(dfmc['total_ar.1'] - dfmc['total_ar']).mean()}\n")



from typing import (
    Optional,
    Union,
)

from fortuna.conformal.multivalid.mixins.classification.binary_multicalibrator import (
    BinaryClassificationMulticalibratorMixin,
)
from fortuna.conformal.multivalid.one_shot.multicalibrator import OneShotMulticalibrator
from fortuna.typing import Array


class OneShotBinaryClassificationMulticalibrator(
    BinaryClassificationMulticalibratorMixin, OneShotMulticalibrator
):
    def __init__(self, seed: int = 0):
        super().__init__(seed=seed)

    def calibrate(
        self,
        targets: Array,
        probs: Optional[Array] = None,
        test_probs: Optional[Array] = None,
        n_buckets: int = 100,
        min_prob_b: Union[float, str] = "auto",
    ):
        return super().calibrate(
            scores=targets,
            values=probs,
            test_values=test_probs,
            n_buckets=n_buckets,
            min_prob_b=min_prob_b,
        )

    def apply_patches(
        self,
        probs: Array,
    ) -> Array:
        return super().apply_patches(values=probs)



from __future__ import annotations

import abc
from typing import (
    Callable,
    Iterable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from flax import jax_utils
import jax
from jax.tree_util import tree_map

from fortuna.data.loader.utils import IterableData
from fortuna.typing import (
    Array,
    Batch,
    InputData,
    Shape,
    Status,
    Targets,
)

T = TypeVar("T")


class BaseDataLoaderABC(abc.ABC):
    def __init__(self, iterable: IterableData, num_unique_labels: int = None):
        self._iterable = iterable
        self._num_unique_labels = num_unique_labels

    def __iter__(self) -> Iterable[Batch]:
        yield from self._iterable

    @property
    def size(self) -> int:
        """
        The number of data points in the data loader.

        Returns
        -------
        int
            Number of data points.
        """
        c = 0
        for inputs, targets in self:
            if isinstance(inputs, dict):
                inputs = inputs[list(inputs.keys())[0]]
            c += inputs.shape[0]
        return c

    @property
    @abc.abstractmethod
    def num_unique_labels(self) -> Optional[int]:
        """
        Number of unique target labels in the task (classification only)

        Returns
        -------
        int
            Number of unique target labels in the task if it is a classification one.
            Otherwise returns None.
        """
        return self._num_unique_labels

    def __len__(self) -> int:
        """
        The number of batches in the data loader.

        Returns
        -------
        int
            Number of batches.
        """
        return sum(1 for _ in self)

    @property
    @abc.abstractmethod
    def input_shape(self) -> Shape:
        """
        Get the shape of the inputs in the data loader.
        """
        pass

    @abc.abstractmethod
    def to_inputs_loader(self) -> BaseInputsLoader:
        """
        Reduce a data loader to an inputs loader.

        Returns
        -------
        BaseInputsLoader
            The inputs loader derived from the data loader. This will be a concrete instance of a subclass
            of :class:`~fortuna.data.loader.BaseInputsLoader`.
        """
        pass

    @abc.abstractmethod
    def to_targets_loader(self) -> BaseTargetsLoader:
        """
        Reduce a data loader to a targets loader.

        Returns
        -------
        BaseTargetsLoader
            The targets loader derived from the data loader. This will be a concrete instance of a subclass
            of :class:`~fortuna.data.loader.BaseTargetsLoader`.
        """
        pass

    @abc.abstractmethod
    def to_array_targets(self) -> Array:
        """
        Reduce a data loader to an array of target data.

        Returns
        -------
        Array
            Array of input data.
        """
        pass

    @classmethod
    def from_callable_iterable(
        cls: Type[T],
        fun: Callable[
            [],
            Iterable[Batch],
        ],
    ) -> T:
        """
        Transform a callable iterable into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseDataLoader`.

        Parameters
        ----------
        fun: Callable[[], Iterable[Batch]]
            A callable iterable of tuples of input and target arrays.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseDataLoader`.
        """
        return cls(iterable=IterableData.from_callable(fun))

    @classmethod
    def from_iterable(
        cls: Type[T],
        iterable: Iterable[Batch],
    ) -> T:
        """
        Transform an iterable into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseDataLoader`.

        Parameters
        ----------
        iterable: Iterable[Batch]
            An iterable of tuples of input and target arrays.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseDataLoader`.
        """
        return cls(iterable=IterableData.from_iterable(iterable))

    @classmethod
    def from_tensorflow_data_loader(cls: Type[T], tf_data_loader) -> T:
        """
        Transform a TensorFlow data loader into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseDataLoader`.

        Parameters
        ----------
        tf_data_loader
            A TensorFlow data loader where each batch is a tuple of input and target Tensors.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseDataLoader`.
        """
        return cls(iterable=IterableData.from_tf_dataloader(tf_data_loader))

    @classmethod
    def from_torch_data_loader(cls: Type[T], torch_data_loader) -> T:
        """
        Transform a PyTorch data loader into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseDataLoader`.

        Parameters
        ----------
        torch_data_loader
            A PyTorch data loader where each batch is a tuple of input and target Tensors.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseDataLoader`.
        """
        return cls(iterable=IterableData.from_torch_dataloader(torch_data_loader))

    @classmethod
    def from_inputs_loaders(
        cls: Type[T],
        inputs_loaders: List[BaseInputsLoader],
        targets: List[int],
        how: str = "interpose",
    ) -> T:
        """
        Transform a list of inputs loader into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseDataLoader`. The newly created data loader is formed out
        of concatenated batches of inputs and the respective assigned target variable.

        Parameters
        ----------
        inputs_loaders: List[BaseInputsLoader]
            A list of inputs loaders.
        targets: List[int]
            A target variable for each inputs loader.
        how: str
            How the input_loaders will be combined: 'interpose' will interpose the input_loaders based on their
            batch sizes; 'concatenate' will ignore batch size and concatenate them.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseDataLoader`.
            The data loader object is formed by the concatenated batches of inputs, and the assigned targets.
        """
        return cls(
            iterable=IterableData.inputs_loaders_to_batch_iterable(
                inputs_loaders=inputs_loaders, targets=targets, how=how
            )
        )

    def to_transformed_data_loader(
        self: T,
        transform: Callable[
            [InputData, Array, Status], Tuple[InputData, Array, Status]
        ],
        status: Optional[Status] = None,
    ) -> T:
        """
        Transform the batches of an existing data loader.

        Parameters
        ----------
        transform : Callable[[InputData, Array, Status], Tuple[InputData, Array, Status]]
            A transformation function. It takes a batch and returns its transformation. A status may be updated
            during the process.
        status : Optional[Status]
            An initial status. This may include pre-computed objects used by the transformation.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseDataLoader`.
        """
        cls = self.__class__
        return cls(
            IterableData.transform_data_loader(
                loader=self, transform_fn=transform, status=status
            )
        )


class BaseInputsLoader:
    def __init__(self, iterable):
        self._iterable = iterable

    def __iter__(self) -> Iterable[InputData]:
        yield from self._iterable

    @property
    def size(self) -> int:
        """
        The number of data points in the inputs loader.

        Returns
        -------
        int
            Number of data points.
        """
        c = 0
        for inputs in self:
            if isinstance(inputs, dict):
                inputs = inputs[list(inputs.keys())[0]]
            c += inputs.shape[0]
        return c

    @property
    def input_shape(self) -> Shape:
        """Get the shape of the inputs in the inputs loader."""

        def fun():
            for inputs in self:
                if isinstance(inputs, dict):
                    input_shape = {k: v.shape[1:] for k, v in inputs.items()}
                else:
                    input_shape = inputs.shape[1:]
                break
            return input_shape

        return fun()

    def __len__(self) -> int:
        """
        The number of batches in the inputs loader.

        Returns
        -------
        int
            Number of batches.
        """
        return sum(1 for _ in self)

    @classmethod
    def from_data_loader(cls: Type[T], data_loader: BaseDataLoaderABC) -> T:
        """
        Reduce a data loader to an inputs loader.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseInputsLoader`.
        """
        return cls(iterable=IterableData.data_loader_to_inputs_iterable(data_loader))

    @classmethod
    def from_callable_iterable(
        cls: Type[T],
        fun: Callable[[], Iterable[InputData]],
    ) -> T:
        """
        Transform a callable iterable into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseInputsLoader`

        Parameters
        ----------
        fun: Callable[[], Iterable[InputData]]
            A callable iterable of input data.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseInputsLoader`.
        """
        return cls(iterable=IterableData.from_callable(fun))

    @classmethod
    def from_iterable(
        cls: Type[T],
        iterable: Iterable[InputData],
    ) -> T:
        """
        Transform an iterable into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseInputsLoader`

        Parameters
        ----------
        iterable: Iterable[InputData]
            An iterable of input data.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseInputsLoader`.
        """
        return cls(iterable=IterableData.from_iterable(iterable))

    def to_transformed_inputs_loader(
        self: T,
        transform: Callable[[InputData, Status], Tuple[InputData, Status]],
        status: Optional[Status] = None,
    ) -> T:
        """
        From an existing loader of inputs, create a loader with transformed inputs.

        Parameters
        ----------
        transform : Callable[[Array, Status], Tuple[Array, Status]]
            A transformation function. It takes a batch of inputs and returns their transformation.
        status : Optional[Status]
            An initial status. This may include pre-computed objects used by the transformation.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseInputsLoader`.
        """
        cls = self.__class__
        return cls(
            IterableData.transform_inputs_or_targets_loader(
                loader=self, transform_fn=transform, status=status
            )
        )


class BaseTargetsLoader:
    """
    A targets loader class.
    """

    def __init__(self, iterable):
        self._iterable = iterable

    def __iter__(self) -> Iterable[Targets]:
        yield from self._iterable

    @property
    def size(self) -> int:
        """
        The number of data points in the targets loader.

        Returns
        -------
        int
            Number of data points.
        """
        c = 0
        for targets in self:
            c += targets.shape[0]
        return c

    def __len__(self) -> int:
        """
        The number of batches in the targets loader.

        Returns
        -------
        int
            Number of batches.
        """
        return sum(1 for _ in self)

    @classmethod
    def from_data_loader(cls: Type[T], data_loader: BaseDataLoaderABC) -> T:
        """
        Reduce a data loader to a targets loader.

        Parameters
        ----------
        data_loader : DataLoader
            A data loader.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseTargetsLoader`.
        """
        return IterableData.data_loader_to_targets_iterable(data_loader)

    @classmethod
    def from_callable_iterable(
        cls: Type[T],
        fun: Callable[[], Iterable[Array]],
    ) -> T:
        """
        Transform a callable iterable into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseTargetsLoader`.

        Parameters
        ----------
        fun: Callable[[], Iterable[Union[Batch, InputData, Array]]],
            A callable iterable of target arrays.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseTargetsLoader`.
        """
        return cls(iterable=IterableData.from_callable(fun))

    @classmethod
    def from_iterable(
        cls: Type[T],
        iterable: Iterable[Array],
    ) -> T:
        """
        Transform an iterable into a concrete instance of a subclass of
        :class:`~fortuna.data.loader.BaseTargetsLoader`.

        Parameters
        ----------
        iterable: Iterable[Array]
            An iterable of target arrays.

        Returns
        -------
        T
           A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseTargetsLoader`.
        """
        return cls(iterable=IterableData.from_iterable(iterable))

    def to_transformed_targets_loader(
        self: T,
        transform: Callable[[Array, Status], Tuple[Array, Status]],
        status: Optional[Status] = None,
    ) -> T:
        """
        From an existing loader of targets, create a loader with transformed targets.

        Parameters
        ----------
        transform : Callable[[Array, Status], Tuple[Array, Status]]
            A transformation function. It takes a batch of targets and returns their transformation. A status may be
            updated during the process.
        status : Optional[Status]
            An initial status. This may include pre-computed objects used by the transformation.

        Returns
        -------
        T
            A concrete instance of a subclass of :class:`~fortuna.data.loader.BaseTargetsLoader`.
        """
        cls = self.__class__
        return cls(
            IterableData.transform_inputs_or_targets_loader(
                loader=self, transform_fn=transform, status=status
            )
        )


class DeviceDimensionAugmentedLoader:
    def __init__(self, loader):
        self._loader = loader

    @staticmethod
    def _reshape_inputs(inputs):
        n_devices = jax.local_device_count()
        if inputs.shape[0] % n_devices != 0:
            raise ValueError(
                f"The size of all batches of inputs must be a multiple of {n_devices}, that is the number of "
                f"available devices. However, a batch of inputs with shape {inputs.shape[0]} was found. "
                f"Please set an appropriate batch size."
            )
        return inputs.reshape((n_devices, -1) + inputs.shape[1:])

    def __iter__(self, *args, **kwargs):
        loader = map(lambda batch: tree_map(self._reshape_inputs, batch), self._loader)
        loader = jax_utils.prefetch_to_device(loader, 2)
        yield from loader


class ConcatenatedLoader:
    def __init__(
        self,
        loaders: List[Any],
    ):
        self._loaders = loaders

    def __iter__(self, *args, **kwargs):
        for loader in self._loaders:
            yield from loader



from typing import Tuple

import jax.numpy as jnp

from fortuna.typing import Array


def scaled_mse_fn(
    outputs: Array,
    targets: Array,
) -> jnp.ndarray:
    """
    Compute a variance-scaled mean-squared-error (MSE).

    Parameters
    ----------
    outputs: Array
        Model outputs to be passed to the loss.
    targets: Array
        Target data points.

    Returns
    -------
    Tuple[jnp.ndarray, Any]
        Scaled MSE evalution and auxiliary objects.
    """
    means, log_vars = jnp.split(outputs, 2, axis=-1)
    return jnp.mean(jnp.sum(jnp.exp(-log_vars) * (targets - means) ** 2 + log_vars, -1))



from typing import Optional

import jax.numpy as jnp

from fortuna.conformal.regression.base import ConformalRegressor
from fortuna.typing import Array


class OneDimensionalUncertaintyConformalRegressor(ConformalRegressor):
    def score(
        self,
        val_preds: Array,
        val_uncertainties: Array,
        val_targets: Array,
    ) -> jnp.ndarray:
        """
        Compute the conformal scores.

        Parameters
        ----------
        val_preds: Array
            A two-dimensional array of predictions over the validation data points.
        val_uncertainties: Array
            A two-dimensional array of uncertainty estimates (e.g. the standard deviation). The first
            dimension is over the validation inputs. The second must have only one component.
        val_targets: Array
            A two-dimensional array of validation target variables.

        Returns
        -------
        jnp.ndarray
            Scores.
        """
        if val_preds.ndim != 2 or val_preds.shape[1] != 1:
            raise ValueError(
                """`val_preds` must be a two-dimensional array. The second dimension must have only one
            component."""
            )
        if val_uncertainties.ndim != 2 or val_uncertainties.shape[1] != 1:
            raise ValueError(
                """`val_uncertainties` must be a two-dimensional array. The second dimension must have only
            one component."""
            )
        if (val_uncertainties <= 0).any():
            raise ValueError(
                """All elements in `val_uncertainties` must be strictly positive."""
            )
        if val_targets.shape[1] != 1:
            raise ValueError(
                """The second dimension of the array(s) in `val_targets` must have only one component."""
            )
        return (jnp.abs(val_targets - val_preds) / val_uncertainties).squeeze(1)

    def quantile(
        self,
        val_preds: Array,
        val_uncertainties: Array,
        val_targets: Array,
        error: float,
        scores: Optional[Array] = None,
    ) -> Array:
        """
        Compute a quantile of the scores.

        Parameters
        ----------
        val_preds: Array
            A two-dimensional array of predictions over the validation data points.
        val_uncertainties: Array
            A two-dimensional array of uncertainty estimates (e.g. the standard deviation). The first
            dimension is over the validation inputs. The second must have only one component.
        val_targets: Array
            A two-dimensional array of validation target variables.
        error: float
            Coverage error. This must be a scalar between 0 and 1, extremes included.
        scores: Optional[float]
            Conformal scores. This should be the output of
            :meth:`~fortuna.conformal.regression.onedim_uncertainty.OneDimensionalUncertaintyConformalRegressor.score`.

        Returns
        -------
        float
            The conformal quantile.
        """
        if error < 0 or error > 1:
            raise ValueError("""`error` must be a scalar between 0 and 1.""")
        if scores is None:
            scores = self.score(val_preds, val_uncertainties, val_targets)
        n = scores.shape[0]
        return jnp.quantile(scores, jnp.ceil((n + 1) * (1 - error)) / n)

    def conformal_interval(
        self,
        val_preds: Array,
        val_uncertainties: Array,
        test_preds: Array,
        test_uncertainties: Array,
        val_targets: Array,
        error: float,
        quantile: Optional[float] = None,
    ) -> jnp.ndarray:
        """
        Coverage interval of each of the test inputs, at the desired coverage error. This is supported only for
        one-dimensional target variables.

        Parameters
        ----------
        val_preds: Array
            A two-dimensional array of predictions over the validation data points.
        test_preds: Array
            A two-dimensional array of predictions over the test data points.
        val_uncertainties: Array
            A two-dimensional array of uncertainty estimates (e.g. the standard deviation). The first
            dimension is over the validation inputs. The second must have only one component.
        test_uncertainties: Array
            A two-dimensional array of uncertainty estimates (e.g. the standard deviation). The first
            dimension is over the test inputs. The second must have only one component.
        val_targets: Array
            A two-dimensional array of validation target variables.
        error: float
            Coverage error. This must be a scalar between 0 and 1, extremes included.
        quantile: Optional[float]
            Conformal quantile. This should be the output of
            :meth:`~fortuna.conformal.regression.onedim_uncertainty.OneDimensionalUncertaintyConformalRegressor.quantile`.

        Returns
        -------
        jnp.ndarray
            The conformal intervals. The two components of the second axis correspond to the left and right interval
            bounds.
        """
        if test_preds.ndim != 2 or test_preds.shape[1] != 1:
            raise ValueError(
                """`test_preds` must be a two-dimensional array. The second dimension must have only one
            component."""
            )
        if test_uncertainties.ndim != 2 or val_uncertainties.shape[1] != 1:
            raise ValueError(
                """`test_uncertainties` must be a two-dimensional array. The second dimension must have
            only one component."""
            )
        if (test_uncertainties <= 0).any():
            raise ValueError(
                """All elements in `test_uncertainties` must be strictly positive."""
            )
        if quantile is None:
            quantile = self.quantile(val_preds, val_uncertainties, val_targets, error)
        lows = test_preds - test_uncertainties * quantile
        highs = test_preds + test_uncertainties * quantile
        return jnp.array(list(zip(lows.squeeze(1), highs.squeeze(1))))



from typing import (
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from jax import (
    ShapeDtypeStruct,
    jit,
    jvp,
    lax,
    pure_callback,
    vjp,
    vmap,
)
import jax.numpy as jnp
from jax.tree_util import (
    tree_map,
    tree_reduce,
)

from fortuna.typing import (
    AnyKey,
    Array,
    InputData,
    Params,
)
from fortuna.utils.freeze import get_paths_with_label
from fortuna.utils.grad import value_and_jacobian_squared_row_norm
from fortuna.utils.nested_dicts import (
    nested_get,
    nested_set,
)


def probit_scaling(
    apply_fn: Callable[[Params, InputData], jnp.ndarray],
    params: Params,
    x: InputData,
    log_var: Union[float, Array],
    has_aux: bool = False,
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None,
    top_k: Optional[int] = None,
) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
    f, scale = value_and_jacobian_squared_row_norm(
        apply_fn, params, x, has_aux=has_aux, freeze_fun=freeze_fun, top_k=top_k
    )

    if has_aux:
        f, aux = f

    f /= 1 + jnp.pi / 8 * jnp.exp(log_var) * scale

    if has_aux:
        return f, aux
    return f


def sequential_probit_scaling(
    apply_fn: Callable[[Params, InputData], jnp.ndarray],
    params: Params,
    x: InputData,
    log_var: Union[float, Array],
    has_aux: bool = False,
    freeze_fun: Optional[Callable[[Tuple[AnyKey, ...], Array], str]] = None,
    top_k: Optional[int] = None,
    memory: Optional[int] = None,
    n_final_tokens: Optional[int] = None,
    stop_gradient: bool = False,
) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Dict]]:
    params = params.unfreeze()

    params_paths = None
    sub_params = None
    if freeze_fun is not None:
        params_paths = tuple(
            get_paths_with_label(
                params, freeze_fun, label=True, allowed_labels=[True, False]
            )
        )
        sub_params = tuple([nested_get(d=params, keys=path) for path in params_paths])

    def set_params(_p):
        if params_paths is None:
            return _p
        return FrozenDict(nested_set(d=params, key_paths=params_paths, objs=_p))

    def _apply_fn(_p, _x, tau):
        _f = apply_fn(set_params(_p), _x)
        if has_aux:
            _f, _ = _f
        _f = _f[0]
        if _f.ndim > 1:
            _f = _f[tau]
        return _f

    f = apply_fn(params, x)
    if has_aux:
        f, aux = f

    if f.ndim > 3:
        raise ValueError("The model outputs can be at most three dimensional.")
    if f.ndim == 2:
        f = f[:, None]

    n_outputs = f.shape[-1]
    seq_length = f.shape[1]
    if n_final_tokens is None:
        n_final_tokens = seq_length
    if n_final_tokens <= 0 or n_final_tokens > seq_length:
        raise ValueError(
            f"`n_final_tokens` must be greater than 0 and cannot be greater than {seq_length}."
        )
    if memory is None:
        memory = n_final_tokens
    if memory <= 0 or memory > n_final_tokens:
        raise ValueError(
            f"`memory` must be greater than 0 and cannot be greater than {n_final_tokens}."
        )

    block_size = top_k if top_k is not None else n_outputs
    tot_size = memory * block_size
    batch_size = f.shape[0]

    indices = None
    if top_k is not None:
        indices = vmap(
            lambda _fx: vmap(lambda _fxtau: jnp.argsort(_fxtau)[-top_k:])(_fx)
        )(f)

    x = x[:, None] if not isinstance(x, dict) else tree_map(lambda v: v[:, None], x)

    def compute_cov(new_tau, prev_tau):
        new_tau -= 1
        prev_tau -= 1

        @vmap
        def _compute_cov(_x, idx):
            new_idx = idx[new_tau] if idx is not None else None
            prev_idx = idx[prev_tau] if idx is not None else None
            size = n_outputs if idx is None else len(prev_idx)

            new_fun = (
                lambda p: _apply_fn(p, _x, new_tau)[new_idx]
                if idx is not None
                else _apply_fn(p, _x, new_tau)
            )
            prev_fun = (
                lambda p: _apply_fn(p, _x, prev_tau)[prev_idx]
                if idx is not None
                else _apply_fn(p, _x, prev_tau)
            )

            J1J2T_op = lambda v: jvp(
                new_fun,
                (sub_params if params_paths is not None else params,),
                vjp(prev_fun, sub_params if params_paths is not None else params)[1](v),
            )[1]

            return vmap(J1J2T_op)(jnp.eye(size)).T

        return jnp.where(
            prev_tau != -1, _compute_cov(x, indices), jnp.empty(block_size)
        )

    init_tau = seq_length - n_final_tokens + 1

    @jit
    def compute_P(new_tau, old_taus):
        P = vmap(lambda tau: compute_cov(new_tau, tau), out_axes=2)(old_taus)
        return P.reshape(P.shape[0], P.shape[1], P.shape[2] * P.shape[3])

    @vmap
    def get_diag(mat):
        return jnp.diag(mat)

    def fun(carry, tau):
        Jinv, old_taus = carry
        S = compute_cov(tau, tau)

        P = compute_P(tau, old_taus)
        M = jnp.matmul(P, Jinv)
        C = S - jnp.matmul(M, P.swapaxes(1, 2))

        M = M[:, :, block_size:]
        Jinv = Jinv[:, block_size:, block_size:]
        Cinv = jnp.linalg.inv(C)
        MtCinv = jnp.matmul(M.swapaxes(1, 2), Cinv)

        Jinv = jnp.concatenate(
            (
                jnp.concatenate((Jinv + jnp.matmul(MtCinv, M), -MtCinv), axis=2),
                jnp.concatenate((-MtCinv.swapaxes(1, 2), Cinv), axis=2),
            ),
            axis=1,
        )

        old_taus = jnp.concatenate((old_taus[1:], jnp.array([tau])))
        return (Jinv, old_taus), get_diag(C)

    def get_diagCs(_params):
        old_taus = jnp.concatenate(
            (
                jnp.zeros(memory - 1, dtype="int32") - 1,
                jnp.array([init_tau], dtype="int32"),
            )
        )
        C = compute_cov(old_taus[-1], old_taus[-1])

        if n_final_tokens > 1:
            Jinv = jnp.linalg.inv(C)
            Jinv = jnp.concatenate(
                (
                    jnp.zeros((batch_size, (memory - 1) * block_size, tot_size)),
                    jnp.concatenate(
                        (
                            jnp.zeros(
                                (batch_size, block_size, (memory - 1) * block_size)
                            ),
                            Jinv,
                        ),
                        axis=2,
                    ),
                ),
                axis=1,
            )

            _, diagCs = lax.scan(
                fun, (Jinv, old_taus), jnp.arange(init_tau + 1, seq_length + 1)
            )
            diagCs = jnp.concatenate(
                (get_diag(C)[:, None], diagCs.swapaxes(0, 1)), axis=1
            )
        else:
            diagCs = get_diag(C)[:, None]

        return diagCs

    diagCs = get_diagCs(params if sub_params is None else sub_params)
    if stop_gradient:
        diagCs = lax.stop_gradient(diagCs)

    if top_k is not None:
        scales = jnp.max(diagCs, axis=2, keepdims=True).repeat(n_outputs, axis=2)
        scales = vmap(
            lambda i: vmap(
                lambda j: scales[i, j]
                .at[indices[i, j]]
                .set(diagCs[i, j, indices[i, j]])
            )(jnp.arange(seq_length - n_final_tokens, seq_length))
        )(jnp.arange(batch_size))
    else:
        scales = diagCs

    f = jnp.concatenate(
        (
            f[:, : seq_length - n_final_tokens],
            f[:, seq_length - n_final_tokens :]
            / (1 + jnp.pi / 8 * jnp.exp(log_var) * scales),
        ),
        axis=1,
    )

    if seq_length == 1:
        f = f[:, 0]

    if has_aux:
        return f, aux
    return f



from typing import (
    Callable,
    Optional,
    Tuple,
    Union,
)

import jax.numpy as jnp

from fortuna.typing import Array


class FitMonitor:
    def __init__(
        self,
        metrics: Optional[
            Tuple[Callable[[jnp.ndarray, Array], Union[float, Array]], ...]
        ] = None,
        early_stopping_patience: int = 0,
        early_stopping_monitor: str = "val_loss",
        early_stopping_min_delta: float = 0.0,
        eval_every_n_epochs: int = 1,
        disable_training_metrics_computation: bool = False,
        verbose: bool = True,
    ):
        """
        An object to configure the monitoring of the posterior fitting.

        Parameters
        ----------
        metrics: Optional[Callable[[jnp.ndarray, Array], Union[float, Array]]]
            Metrics to monitor during training.
        early_stopping_patience: int
            Number of consecutive epochs without an improvement in the performance on the validation set before stopping
             the training.
        early_stopping_monitor: str
            Validation metric to be monitored for early stopping.
        early_stopping_min_delta: float
            Minimum change between updates to be considered an improvement, i.e., if the absolute change is less than
            `early_stopping_min_delta` then this is not considered an improvement leading to a potential early stop.
        eval_every_n_epochs: int
            Number of training epochs between validation. To disable, set
            `eval_every_n_epochs` to None or 0 (i.e., no validation metrics will be computed during training).
        disable_training_metrics_computation: bool
            if True, during training the only metric computed is the objective function.
            Otherwise, all the metrics provided by the user at runtime will be computed for the training step.
        verbose: bool
            Whether to log the training progress.
        """
        if metrics is not None:
            if type(metrics) != tuple:
                raise ValueError("`metrics` must be a tuple of callable metrics.")
            for metric in metrics:
                if not callable(metric):
                    raise ValueError(
                        f"All metrics in `metrics` must be callable objects, but {metric} is not."
                    )

        self.metrics = metrics
        self.early_stopping_patience = early_stopping_patience
        self.early_stopping_monitor = early_stopping_monitor
        self.early_stopping_min_delta = early_stopping_min_delta
        self.eval_every_n_epochs = eval_every_n_epochs
        self.disable_training_metrics_computation = disable_training_metrics_computation
        self.verbose = verbose



from typing import (
    Callable,
    Dict,
    Optional,
)

from flax import traverse_util
from flax.core import FrozenDict
import jax.numpy as jnp
import optax
from optax._src import base

from fortuna.typing import Params


def stepped_schedule(
    init_value: float, boundaries_and_scales: Optional[Dict[int, float]] = None
) -> base.Schedule:
    """
    Returns a stepped learning rate schedule with optional warmup.

    A stepped learning rate schedule decreases the learning rate
    by specified amounts at specified epochs.
    A common ImageNet schedule decays the learning rate by a factor of 0.1 at epochs 30, 60 and 80.
    This would be specified setting `boundaries_and_scales` to:

      {
        30: 0.1,
        60: 0.01,
        80: 0.001
      }

    Parameters
    ----------
    init_value: float
        An initial value `init_v`.
    boundaries_and_scales: Optional[Dict[int, float]]
        The schedule as a dict like `{epoch_x: lr_factor_x, epoch_y: lr_factor_y}`;
    the step occurs at epoch `epoch` and sets the learning rate to `init_value * lr_factor`

    Returns
    -------
    base.Schedule
        A function that maps step counts to values.
    """
    if boundaries_and_scales is not None:
        all_positive = all(scale >= 0.0 for scale in boundaries_and_scales.values())
        if not all_positive:
            raise ValueError(
                "`piecewise_constant_schedule` expects non-negative scale factors"
            )

    def schedule(count):
        v = init_value
        if boundaries_and_scales is not None:
            for threshold, scale in sorted(boundaries_and_scales.items()):
                indicator = jnp.maximum(0.0, jnp.sign(threshold - count))
                v = v * indicator + (1 - indicator) * scale * init_value
        return v

    return schedule


def linear_scheduler_with_warmup(
    learning_rate: float,
    num_inputs_train: int,
    train_total_batch_size: int,
    num_train_epochs: int,
    num_warmup_steps: int,
) -> Callable[[int], jnp.array]:
    """
    Create a linear scheduler with a warmup

    Parameters
    ----------
    learning_rate: float
        Learning rate value
    num_inputs_train: int
        Number of input data points in the training set.
    train_total_batch_size: int
        Total training batch size.
    num_train_epochs: int
        Number of training epochs.
    num_warmup_steps: int
        Number of warmup steps. At the end of the warm-up phase the learning rate value will be `learning_rate`.

    Returns
    -------
    Callable[[int], jnp.array]
        The scheduler function
    """
    steps_per_epoch = num_inputs_train // train_total_batch_size
    num_train_steps = steps_per_epoch * num_train_epochs
    warmup_fn = optax.linear_schedule(
        init_value=0.0, end_value=learning_rate, transition_steps=num_warmup_steps
    )
    decay_fn = optax.linear_schedule(
        init_value=learning_rate,
        end_value=0,
        transition_steps=num_train_steps - num_warmup_steps,
    )
    schedule_fn = optax.join_schedules(
        schedules=[warmup_fn, decay_fn], boundaries=[num_warmup_steps]
    )
    return schedule_fn


def decay_mask_without_layer_norm_fn(params: Params) -> Params:
    flat_params = traverse_util.flatten_dict(params)
    # find out all LayerNorm parameters
    layer_norm_candidates = ["layernorm", "layer_norm", "ln"]
    layer_norm_named_params = {
        layer[-2:]
        for layer_norm_name in layer_norm_candidates
        for layer in flat_params.keys()
        if layer_norm_name in "".join(layer).lower()
    }
    flat_mask = {
        path: (path[-1] != "bias" and path[-2:] not in layer_norm_named_params)
        for path in flat_params
    }
    return FrozenDict(traverse_util.unflatten_dict(flat_mask))



from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
    Union,
)

from flax.core import FrozenDict
import jax.numpy as jnp

from fortuna.training.train_state import TrainState
from fortuna.typing import (
    CalibMutable,
    CalibParams,
    OptaxOptimizer,
)
from fortuna.utils.strings import convert_string_to_jnp_array


class OutputCalibState(TrainState):
    params: CalibParams
    mutable: Optional[CalibMutable] = None
    encoded_name: jnp.ndarray = convert_string_to_jnp_array("OutputCalibState")

    @classmethod
    def init(
        cls,
        params: CalibParams,
        mutable: Optional[CalibMutable] = None,
        optimizer: Optional[OptaxOptimizer] = None,
        **kwargs,
    ) -> Any:
        """
        Initialize an output calibration state.

        Parameters
        ----------
        params : CalibParams
            The calibration parameters.
        optimizer : Optional[OptaxOptimizer]
            An Optax optimizer associated with the calibration state.
        mutable : Optional[CalibMutable]
            The calibration mutable objects.

        Returns
        -------
        Any
            A calibration state.
        """
        return cls(
            apply_fn=None,
            params=params,
            opt_state=kwargs["opt_state"]
            if optimizer is None and "opt_state" in kwargs
            else None
            if optimizer is None
            else optimizer.init(params),
            mutable=mutable,
            step=kwargs.get("step", 0),
            tx=optimizer,
            **{
                k: v
                for k, v in kwargs.items()
                if k not in ["opt_state", "apply_fn", "tx", "step"]
            },
        )

    @classmethod
    def init_from_dict(
        cls,
        d: Union[Dict, FrozenDict],
        optimizer: Optional[OptaxOptimizer] = None,
        **kwargs,
    ) -> OutputCalibState:
        """
        Initialize a calibration state from a dictionary.

        Parameters
        ----------
        d : Union[Dict, FrozenDict]
            A dictionary with as keys the calibrators and as values their initializations.
        optimizer : Optional[OptaxOptimizer]
            An optax optimizer to assign to the calibration state.

        Returns
        -------
        OutputCalibState
            A calibration state.
        """
        kwargs = {
            **kwargs,
            **{
                k: v
                for k, v in d.items()
                if k
                not in [
                    "params",
                    "mutable",
                    "optimizer",
                ]
            },
        }
        return cls.init(
            FrozenDict(d["params"]),
            FrozenDict(d["mutable"]),
            optimizer,
            **kwargs,
        )



from fortuna.prob_model.calib_config.checkpointer import CalibCheckpointer
from fortuna.prob_model.calib_config.monitor import CalibMonitor
from fortuna.prob_model.calib_config.optimizer import CalibOptimizer
from fortuna.prob_model.calib_config.processor import CalibProcessor


class CalibConfig:
    def __init__(
        self,
        optimizer: CalibOptimizer = CalibOptimizer(),
        checkpointer: CalibCheckpointer = CalibCheckpointer(),
        monitor: CalibMonitor = CalibMonitor(),
        processor: CalibProcessor = CalibProcessor(),
    ):
        """
        Configure the probabilistic model calibration.

        Parameters
        ----------
        optimizer: CalibOptimizer
            It defines the optimization specifics.
        checkpointer: CalibCheckpointer
            It handles saving and restoring checkpoints.
        monitor: CalibMonitor
            It monitors training progress and might induce early stopping.
        processor: CalibProcessor
            It processes where computation takes place.
        """
        self.optimizer = optimizer
        self.checkpointer = checkpointer
        self.monitor = monitor
        self.processor = processor



from fortuna.prob_model.posterior.base import PosteriorApproximator
from fortuna.prob_model.posterior.swag import SWAG_NAME


class SWAGPosteriorApproximator(PosteriorApproximator):
    def __init__(self, rank: int = 5):
        """
        SWAG posterior approximator. It is responsible to define how the posterior distribution is approximated.

        Parameters
        ----------
        rank: int
            SWAG approximates the posterior with a Gaussian distribution. The Gaussian's covariance matrix is formed by
            a diagonal matrix, and a low-rank empirical approximation. This argument defines the rank of the low-rank
            empirical covariance approximation. It must be at least 2.
        """
        self.rank = rank

    def __str__(self):
        return SWAG_NAME



from typing import (
    Optional,
    Type,
)

from flax.training.dynamic_scale import DynamicScale
from jax import numpy as jnp


class HashableMixin:
    def __hash__(self) -> int:
        return hash(
            tuple(
                [
                    getattr(self, k)
                    for k in sorted(vars(self).keys())
                    if not k.startswith("_")
                ]
            )
        )

    def __eq__(self, other) -> bool:
        self_keys = [k for k in vars(self).keys() if not k.startswith("_")]
        other_keys = [k for k in vars(other).keys() if not k.startswith("_")]

        same_keys = self_keys == other_keys
        if same_keys and isinstance(other, self.__class__):
            same_vals = all(
                map(lambda k: getattr(self, k) == getattr(other, k), self_keys)
            )
            return same_vals
        return False


def get_dynamic_scale_instance_from_model_dtype(dtype: Type) -> Optional[DynamicScale]:
    if dtype in [jnp.float16, jnp.bfloat16]:
        return DynamicScale()



from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)

from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp

from fortuna.output_calib_model.predictive.base import Predictive
from fortuna.typing import (
    Array,
    CalibMutable,
    CalibParams,
    Outputs,
    Targets,
)
from fortuna.utils.random import WithRNG


class Loss(WithRNG):
    def __init__(
        self, predictive: Predictive, loss_fn: Callable[[Outputs, Targets], jnp.ndarray]
    ):
        self.predictive = predictive
        self.loss_fn = loss_fn

    def __call__(
        self,
        params: CalibParams,
        targets: Array,
        outputs: Array,
        mutable: Optional[CalibMutable] = None,
        rng: Optional[PRNGKeyArray] = None,
        return_aux: Optional[List[str]] = None,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, Any]]:
        if return_aux is None:
            return_aux = []
        supported_aux = ["outputs", "mutable"]
        unsupported_aux = [s for s in return_aux if s not in supported_aux]
        if sum(unsupported_aux) > 0:
            raise AttributeError(
                """The auxiliary objects {} is unknown. Please make sure that all elements of `return_aux`
                            belong to the following list: {}""".format(
                    unsupported_aux, supported_aux
                )
            )
        aux = dict()
        outs = self.predictive.output_calib_manager.apply(
            params=params["output_calibrator"],
            outputs=outputs,
            mutable=mutable["output_calibrator"],
            rng=rng,
            calib="mutable" in return_aux,
        )
        if (
            mutable is not None
            and mutable["output_calibrator"] is not None
            and "mutable" in return_aux
        ):
            outputs, aux["mutable"] = outs
            aux["mutable"] = dict(output_calibrator=aux["mutable"])
        else:
            outputs = outs
            if "mutable" in return_aux:
                aux["mutable"] = dict(output_calibrator=None)

        if "outputs" in return_aux:
            aux["outputs"] = outputs

        loss = self.loss_fn(outputs, targets)

        if len(return_aux) == 0:
            return loss
        else:
            return loss, aux



import importlib
import logging
from typing import (
    Callable,
    Optional,
    Type,
)

from flax import linen as nn
import jax.numpy as jnp

from fortuna.calib_model.base import CalibModel
from fortuna.calib_model.config.base import Config
from fortuna.calib_model.predictive.classification import ClassificationPredictive
from fortuna.data import DataLoader
from fortuna.likelihood.classification import ClassificationLikelihood
from fortuna.loss.classification.focal_loss import focal_loss_fn
from fortuna.model.model_manager.classification import ClassificationModelManager
from fortuna.model_editor.base import ModelEditor
from fortuna.prob_output_layer.classification import (
    ClassificationMaskedProbOutputLayer,
    ClassificationProbOutputLayer,
)
from fortuna.typing import (
    Outputs,
    Status,
    Targets,
)
from fortuna.utils.data import get_input_shape


class CalibClassifier(CalibModel):
    def __init__(
        self,
        model: nn.Module,
        model_editor: Optional[ModelEditor] = None,
        seed: int = 0,
    ):
        r"""
        A calibration classifier class.

        Parameters
        ----------
        model : nn.Module
            A model describing the deterministic relation between inputs and outputs. The outputs must correspond to
            the logits of a softmax probability vector. The output dimension must be the same as the number of classes.
            Let :math:`x` be input variables and :math:`w` the random model parameters. Then the model is described by
            a function :math:`f(w, x)`, where each component of :math:`f` corresponds to one of the classes.
        model_editor : ModelEditor
            A model_editor objects. It takes the forward pass and transforms the outputs.
        seed: int
            A random seed.

        Attributes
        ----------
        model : nn.Module
            See `model` in `Parameters`.
        model_manager : ClassificationModelManager
            This object orchestrates the model's forward pass.
        prob_output_layer : ClassificationProbOutputLayer
            This object characterizes the distribution of target variable given the outputs. It is defined
            by :math:`p(y|o)=\text{Categorical}(y|p=softmax(o))`,
            where :math:`o` denotes the outputs and :math:`y` denotes a target variable.
        likelihood : ClassificationLikelihood
            The likelihood function. This is defined by
            :math:`p(y|w, \phi, x) = \text{Categorical}(y|p=\text{softmax}(g(\phi, f(w, x)))`.
        predictive : ClassificationPredictive
            This denotes the predictive distribution, that is :math:`p(y|x, \mathcal{D})`.
        """
        self.model_manager = self._get_model_manager(
            model, model_editor, ClassificationModelManager
        )
        self.prob_output_layer = self._get_prob_output_layer(model)
        self.likelihood = ClassificationLikelihood(
            model_manager=self.model_manager,
            prob_output_layer=self.prob_output_layer,
            output_calib_manager=None,
        )
        self.predictive = ClassificationPredictive(likelihood=self.likelihood)
        super().__init__(seed=seed)

    def _get_prob_output_layer(self, model: nn.Module) -> ClassificationProbOutputLayer:
        try:
            # import modules if available
            transformers_flax_auto_module = importlib.import_module(
                "transformers.models.auto.modeling_flax_auto"
            )
            FLAX_MODEL_FOR_MASKED_LM_MAPPING_NAMES = list(
                getattr(
                    transformers_flax_auto_module,
                    "FLAX_MODEL_FOR_MASKED_LM_MAPPING_NAMES",
                ).values()
            )
            if str(model.__class__.__name__) in FLAX_MODEL_FOR_MASKED_LM_MAPPING_NAMES:
                prob_output_layer = ClassificationMaskedProbOutputLayer()
            else:
                prob_output_layer = ClassificationProbOutputLayer()
        except ModuleNotFoundError:
            prob_output_layer = ClassificationProbOutputLayer()
        return prob_output_layer

    def _get_model_manager(
        self,
        model: nn.Module,
        model_editor: Optional[nn.Module],
        model_manager_cls: Type,
    ) -> ClassificationModelManager:
        try:
            # import modules if available
            transformers_module = importlib.import_module("transformers")
            fortuna_transformers_classification_module = importlib.import_module(
                "fortuna.model.model_manager.transformers.classification"
            )
            # import relevant classes
            FlaxPreTrainedModel = getattr(transformers_module, "FlaxPreTrainedModel")
            HuggingFaceClassificationModelManager = getattr(
                fortuna_transformers_classification_module,
                "HuggingFaceClassificationModelManager",
            )
            # load model manager
            if isinstance(model, FlaxPreTrainedModel):
                model_manager = HuggingFaceClassificationModelManager(
                    model, model_editor
                )
            else:
                model_manager = model_manager_cls(model, model_editor)
        except ModuleNotFoundError as e:
            logging.warning(
                "No module named 'transformer' is installed. "
                "If you are not working with models from the `transformers` library ignore this warning, otherwise "
                "please install the optional 'transformers' dependency of fortuna."
                'Using poetry, you can achieve this by entering: `poetry install --extras "transformers"`'
            )
            model_manager = model_manager_cls(model, model_editor)
        return model_manager

    def _check_output_dim(self, data_loader: DataLoader):
        if data_loader.size == 0:
            raise ValueError(
                """`data_loader` is either empty or incorrectly constructed."""
            )
        output_dim = data_loader.num_unique_labels
        for x, y in data_loader:
            input_shape = get_input_shape(x)
            break
        model_manager_output_dim = self._get_output_dim(input_shape)
        if model_manager_output_dim != output_dim:
            raise ValueError(
                f"""The outputs dimension of `model` must correspond to the number of different classes
            in the target variables of `data_loader`. However, {model_manager_output_dim} and {output_dim} were
            found, respectively."""
            )

    def calibrate(
        self,
        calib_data_loader: DataLoader,
        val_data_loader: Optional[DataLoader] = None,
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray] = focal_loss_fn,
        config: Config = Config(),
    ) -> Status:
        """
        Calibrate the calibration classifier.

        Parameters
        ----------
        calib_data_loader : DataLoader
            A calibration data loader.
        val_data_loader : DataLoader
            A validation data loader.
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray]
            The loss function to use for calibration.
        config : Config
            An object to configure the calibration.

        Returns
        -------
        Status
            A calibration status object. It provides information about the calibration.
        """
        self._check_output_dim(calib_data_loader)
        if val_data_loader is not None:
            self._check_output_dim(val_data_loader)
        return self._calibrate(
            calib_data_loader=calib_data_loader,
            uncertainty_fn=config.monitor.uncertainty_fn
            if config.monitor.uncertainty_fn is not None
            else self.prob_output_layer.mean,
            val_data_loader=val_data_loader,
            loss_fn=loss_fn,
            config=config,
        )



from __future__ import annotations

import abc
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from flax.core import FrozenDict
from jax import (
    random,
    vmap,
)
from jax._src.prng import PRNGKeyArray
import jax.numpy as jnp
from jax.tree_util import tree_map
from optax._src.base import PyTree

from fortuna.distribution.base import Distribution
from fortuna.prob_model.posterior.posterior_trainer import PosteriorTrainerABC
from fortuna.prob_model.posterior.state import PosteriorState
from fortuna.training.callback import Callback
from fortuna.typing import (
    Array,
    Batch,
    CalibMutable,
    CalibParams,
    Mutable,
    Params,
    Path,
)
from fortuna.utils.strings import encode_tuple_of_lists_of_strings_to_numpy


class NormalizingFlowTrainer(PosteriorTrainerABC):
    @abc.abstractmethod
    def __str__(self):
        pass

    def sample_forward(self, rng: jnp.array, params: any, n_samples: int) -> tuple:
        """
        Sample from the push-forward distribution.

        Parameters
        ----------
        rng: PRNGKeyArray
            Random number generator.
        params: Params
            Transformation parameters.
        n_samples: int
            Number of samples to draw.

        Returns
        -------
        tuple
            forward_samples: jnp.ndarray
                Samples.
            ldj: jnp.ndarray
                Log-determinant of the Jacobians.
        """
        return self.forward(params, self.sample_base(rng, n_samples))

    def __init__(
        self,
        *,
        base: Distribution,
        architecture: object,
        which_params: Optional[Tuple[List[str]]],
        unravel: Callable,
        sub_unravel: Callable,
        **kwargs,
    ):
        super(NormalizingFlowTrainer, self).__init__(**kwargs)
        # base distribution
        self.sample_base = base.sample
        self.base_log_joint_prob = base.log_joint_prob

        # architecture
        self.forward = architecture.forward
        self.backward = architecture.backward
        self.init_params = architecture.init_params

        # Normalizing flows on subsets of the model parameters
        self._which_params = which_params
        self._encoded_which_params = encode_tuple_of_lists_of_strings_to_numpy(
            which_params
        )
        self._unravel = unravel
        self._sub_unravel = sub_unravel

    def training_loss_step(
        self,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        params: Params,
        batch: Batch,
        mutable: Mutable,
        rng: PRNGKeyArray,
        n_data: int,
        unravel: Optional[Callable[[any], PyTree]] = None,
        calib_params: Optional[CalibParams] = None,
        calib_mutable: Optional[CalibMutable] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[jnp.ndarray, Dict[str, Any]]:
        return_aux = ["outputs"]
        if mutable is not None:
            return_aux += ["mutable"]
        rng, key = random.split(rng)
        v, ldj = self.sample_forward(key, params, kwargs["n_samples"])
        neg_logp, aux = vmap(
            lambda _rav_params: loss_fun(
                unravel(_rav_params),
                batch,
                n_data=n_data,
                mutable=mutable,
                return_aux=return_aux,
                train=True,
                rng=rng,
                calib_params=calib_params,
                calib_mutable=calib_mutable,
            )
        )(v)
        return (
            jnp.mean(neg_logp) - jnp.mean(ldj),
            {
                "outputs": aux.get("outputs"),
                "mutable": tree_map(lambda x: x[-1], aux.get("mutable")),
                "logging_kwargs": dict(),
            },
        )

    def training_step_end(
        self,
        current_epoch: int,
        state: PosteriorState,
        aux: Dict[str, Any],
        batch: Batch,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float], ...]] = None,
        callbacks: Optional[List[Callback]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Tuple[PosteriorState, Dict[str, jnp.ndarray]]:
        if (
            self.save_checkpoint_dir is not None
            and self.save_every_n_steps is not None
            and self.save_every_n_steps > 0
            and self._global_training_step >= self.save_every_n_steps
            and self._global_training_step % self.save_every_n_steps == 0
        ):
            self.save_checkpoint(
                state, self.save_checkpoint_dir, keep=self.keep_top_n_checkpoints
            )
        training_losses_and_metrics = {"loss": aux["loss"]}

        if aux["logging_kwargs"] is not None:
            for k, v in aux["logging_kwargs"].items():
                training_losses_and_metrics[k] = v

        if not self.disable_training_metrics_computation and metrics is not None:
            preds = self.predict_fn(aux["outputs"])
            if self.multi_device:
                training_batch_metrics = vmap(
                    lambda p: self.compute_metrics(
                        p,
                        batch[1].reshape(
                            (batch[1].shape[0] * batch[1].shape[1],)
                            + batch[1].shape[2:]
                        ),
                        metrics,
                    )
                )(
                    preds.reshape(
                        (
                            preds.shape[1],
                            preds.shape[0] * preds.shape[2],
                        )
                        + preds.shape[3:]
                    )
                )
            else:
                training_batch_metrics = vmap(
                    lambda p: self.compute_metrics(p, batch[1], metrics)
                )(preds)
            training_batch_metrics = tree_map(
                lambda m: m.mean(), training_batch_metrics
            )
            for k, v in training_batch_metrics.items():
                training_losses_and_metrics[k] = v
        state = self._callback_loop(state, callbacks, "training_step_end")
        return state, training_losses_and_metrics

    def validation_step(
        self,
        state: PosteriorState,
        batch: Batch,
        loss_fun: Callable[[Any], Union[float, Tuple[float, dict]]],
        rng: PRNGKeyArray,
        n_data: int,
        metrics: Optional[Tuple[Callable[[jnp.ndarray, Array], float]], ...] = None,
        unravel: Optional[Callable[[any], PyTree]] = None,
        kwargs: FrozenDict[str, Any] = FrozenDict(),
    ) -> Dict[str, jnp.ndarray]:
        rng, key = random.split(rng)
        v, ldj = self.sample_forward(key, state.params, kwargs["n_samples"])
        neg_logp, aux = vmap(
            lambda _rav_params: loss_fun(
                self._unravel(_rav_params),
                batch,
                n_data=n_data,
                mutable=state.mutable,
                return_aux=["outputs"],
                train=False,
                rng=rng,
                calib_params=state.calib_params,
                calib_mutable=state.calib_mutable,
            )
        )(v)
        loss = jnp.mean(neg_logp) - jnp.mean(ldj)
        if metrics is not None:
            preds = self.predict_fn(aux["outputs"])
            val_metrics = vmap(lambda p: self.compute_metrics(p, batch[1], metrics))(
                preds
            )
            val_metrics = tree_map(lambda m: m.mean(), val_metrics)
            return {
                "val_loss": loss,
                **{f"val_{m}": v for m, v in val_metrics.items()},
            }
        return dict(val_loss=loss)



from fortuna.typing import Array


class ConformalRegressor:
    """
    A base conformal regressor class.
    """

    def is_in(self, values: Array, conformal_intervals: Array) -> Array:
        """
        Check whether the values lie within their respective conformal intervals.

        Parameters
        ----------
        values: Array
            Values to check if they lie in the respective conformal intervals.
        conformal_intervals: Array
            A conformal interval for each input data point.

        Returns
        -------
        Array
            An array of ones or zero, indicating whether the values lie within their respective conformal intervals.
        """
        return (values <= conformal_intervals[:, 1]) * (
            values >= conformal_intervals[:, 0]
        )



from typing import Optional

import jax.numpy as jnp

from fortuna.conformal.regression.base import ConformalRegressor
from fortuna.typing import Array


class QuantileConformalRegressor(ConformalRegressor):
    def score(
        self,
        val_lower_bounds: Array,
        val_upper_bounds: Array,
        val_targets: Array,
    ) -> jnp.ndarray:
        """
        Compute score function.

        Parameters
        ----------
        val_lower_bounds: Array
            Interval lower bounds computed on a validation set.
        val_upper_bounds: Array
            Interval upper bounds computed on a validation set.
        val_targets: Array
            A two-dimensional array of validation target variables.

        Returns
        -------
        jnp.ndarray
            The conformal scores.
        """
        if val_lower_bounds.shape != val_upper_bounds.shape:
            raise ValueError(
                """The shapes of `val_lower_bounds` and `val_upper_bounds` must be the same, but shapes {} and {}
                were given, respectively.""".format(
                    val_lower_bounds.shape, val_upper_bounds.shape
                )
            )
        if val_lower_bounds.ndim != 1:
            raise ValueError(
                "`val_lower_bounds` and `val_upper_bounds` must be one-dimensional arrays."
            )
        if val_targets.shape[1] != 1:
            raise ValueError(
                """The second dimension of `val_targets` must have only one component."""
            )
        val_targets = val_targets.squeeze(1)
        return jnp.maximum(
            val_lower_bounds - val_targets, val_targets - val_upper_bounds
        )

    def quantile(
        self,
        val_lower_bounds: Array,
        val_upper_bounds: Array,
        val_targets: Array,
        error: float,
        scores: Optional[Array] = None,
    ) -> Array:
        """
        Compute a quantile of the scores.

        Parameters
        ----------
        val_lower_bounds: Array
            Interval lower bounds computed on a validation set.
        val_upper_bounds: Array
            Interval upper bounds computed on a validation set.
        val_targets: Array
            A two-dimensional array of validation target variables.
        error: float
            Coverage error. This must be a scalar between 0 and 1, extremes included. This should correspond to the
            coverage error for which `val_lower_bounds`, `val_upper_bounds`, `test_lower_bounds` and
            `test_upper_bounds` were computed.
        scores: Optional[float]
            Conformal scores. This should be the output of
            :meth:`~fortuna.conformal.regression.quantile.QuantileConformalRegressor.score`.

        Returns
        -------
        float
            The conformal quantile.
        """
        if error < 0 or error > 1:
            raise ValueError("""`error` must be a scalar between 0 and 1.""")
        if scores is None:
            scores = self.score(val_lower_bounds, val_upper_bounds, val_targets)
        n = scores.shape[0]
        return jnp.quantile(scores, jnp.ceil((n + 1) * (1 - error)) / n)

    def conformal_interval(
        self,
        val_lower_bounds: Array,
        val_upper_bounds: Array,
        test_lower_bounds: Array,
        test_upper_bounds: Array,
        val_targets: Array,
        error: float,
        quantile: Optional[float] = None,
    ) -> jnp.ndarray:
        """
        Coverage interval of each of the test inputs, at the desired coverage error. This is supported only for
        scalar target variables.

        Parameters
        ----------
        val_lower_bounds: Array
            Interval lower bounds computed on a validation set.
        val_upper_bounds: Array
            Interval upper bounds computed on a validation set.
        test_lower_bounds: Array
            Interval lower bounds computed on a test set.
        test_upper_bounds: Array
            Interval upper bounds computed on a test set.
        val_targets: Array
            A two-dimensional array of validation target variables.
        error: float
            Coverage error. This must be a scalar between 0 and 1, extremes included. This should correspond to the
            coverage error for which `val_lower_bounds`, `val_upper_bounds`, `test_lower_bounds` and
            `test_upper_bounds` were computed.
        quantile: Optional[float]
            Conformal quantiles. This should be the output of
            :meth:`~fortuna.conformal.regression.quantile.QuantileConformalRegressor.quantile`.

        Returns
        -------
        jnp.ndarray
            The conformal intervals. The two components of the second axis correspond to the left and right interval
            bounds.
        """
        if val_lower_bounds.shape != val_upper_bounds.shape:
            raise ValueError(
                f"""The shapes of `val_lower_bounds` and `val_upper_bounds` must be the same, but shapes
                {val_lower_bounds.shape} and {val_upper_bounds.shape} were found, respectively."""
            )

        if test_lower_bounds.shape != test_upper_bounds.shape:
            raise ValueError(
                f"""The shapes of `test_lower_bounds` and `test_upper_bounds` must be the same, but shapes
                {test_lower_bounds.shape} and {test_upper_bounds.shape} were found, respectively."""
            )

        if val_lower_bounds.ndim not in [1, 2]:
            raise ValueError(
                "`val_lower_bounds` and `val_upper_bounds` must be one- or two-dimensional arrays. If "
                "two-dimensional, the second dimension must be 1."
            )
        if test_lower_bounds.ndim not in [1, 2]:
            raise ValueError(
                "`test_lower_bounds` and `test_upper_bounds` must be one- or two-dimensional arrays. If "
                "two-dimensional, the second dimension must be 1."
            )
        if val_lower_bounds.ndim == 2:
            if val_lower_bounds.shape[1] != 1:
                raise ValueError(
                    f"The second dimension of `val_lower_bounds` must have only one component, but"
                    f"{val_lower_bounds.shape[1]} components were found."
                )
            else:
                val_lower_bounds = val_lower_bounds.squeeze(1)
                val_upper_bounds = val_upper_bounds.squeeze(1)
        if test_lower_bounds.ndim == 2:
            if test_lower_bounds.shape[1] != 1:
                raise ValueError(
                    f"The second dimension of `test_lower_bounds` must have only one component, but"
                    f"{test_lower_bounds.shape[1]} components were found."
                )
            else:
                test_lower_bounds = test_lower_bounds.squeeze(1)
                test_upper_bounds = test_upper_bounds.squeeze(1)

        if quantile is None:
            quantile = self.quantile(
                val_lower_bounds, val_upper_bounds, val_targets, error
            )
        lows = test_lower_bounds - quantile
        highs = test_upper_bounds + quantile
        return jnp.array(list(zip(lows, highs)))



import abc
import logging
from typing import (
    Callable,
    Optional,
)

from flax.core import FrozenDict
import jax.numpy as jnp

from fortuna.output_calib_model.config.base import Config
from fortuna.output_calib_model.loss import Loss
from fortuna.output_calib_model.output_calib_mixin import (
    WithOutputCalibCheckpointingMixin,
)
from fortuna.output_calib_model.output_calib_model_calibrator import (
    JittedOutputCalibModelCalibrator,
    MultiDeviceOutputCalibModelCalibrator,
    OutputCalibModelCalibrator,
)
from fortuna.output_calib_model.output_calib_state_repository import (
    OutputCalibStateRepository,
)
from fortuna.output_calib_model.state import OutputCalibState
from fortuna.output_calibrator.output_calib_manager.state import OutputCalibManagerState
from fortuna.typing import (
    Array,
    Outputs,
    Path,
    Status,
    Targets,
)
from fortuna.utils.device import select_trainer_given_devices
from fortuna.utils.random import RandomNumberGenerator


class OutputCalibModel(WithOutputCalibCheckpointingMixin, abc.ABC):
    """
    Abstract calibration model class.
    """

    def __init__(self, seed: int = 0):
        super().__init__()
        self.rng = RandomNumberGenerator(seed=seed)
        self.__set_rng()

    def __set_rng(self):
        self.output_calib_manager.rng = self.rng
        self.prob_output_layer.rng = self.rng
        self.predictive.rng = self.rng

    def _calibrate(
        self,
        uncertainty_fn: Callable[[jnp.ndarray, jnp.ndarray, Array], jnp.ndarray],
        loss_fn: Callable[[Outputs, Targets], jnp.ndarray],
        calib_outputs: Array,
        calib_targets: Array,
        val_outputs: Optional[Array] = None,
        val_targets: Optional[Array] = None,
        config: Config = Config(),
    ) -> Status:
        if (val_targets is not None and val_outputs is None) or (
            val_targets is None and val_outputs is not None
        ):
            raise ValueError(
                "For validation, both `val_outputs` and `val_targets` must be passed as arguments."
            )
        trainer_cls = select_trainer_given_devices(
            devices=config.processor.devices,
            base_trainer_cls=OutputCalibModelCalibrator,
            jitted_trainer_cls=JittedOutputCalibModelCalibrator,
            multi_device_trainer_cls=MultiDeviceOutputCalibModelCalibrator,
            disable_jit=config.processor.disable_jit,
        )

        calibrator = trainer_cls(
            calib_outputs=calib_outputs,
            calib_targets=calib_targets,
            val_outputs=val_outputs,
            val_targets=val_targets,
            predict_fn=self.prob_output_layer.predict,
            uncertainty_fn=uncertainty_fn,
            save_checkpoint_dir=config.checkpointer.save_checkpoint_dir,
            save_every_n_steps=config.checkpointer.save_every_n_steps,
            keep_top_n_checkpoints=config.checkpointer.keep_top_n_checkpoints,
            disable_training_metrics_computation=config.monitor.disable_calibration_metrics_computation,
            eval_every_n_epochs=config.monitor.eval_every_n_epochs,
            early_stopping_monitor=config.monitor.early_stopping_monitor,
            early_stopping_min_delta=config.monitor.early_stopping_min_delta,
            early_stopping_patience=config.monitor.early_stopping_patience,
        )

        if config.checkpointer.restore_checkpoint_path is None:
            state = OutputCalibManagerState.init_from_dict(
                d=FrozenDict(
                    output_calibrator=self.output_calib_manager.init(
                        output_dim=calib_outputs.shape[-1]
                    )
                ),
            )
            state = OutputCalibState.init(
                params=state.params,
                mutable=state.mutable,
                optimizer=config.optimizer.method,
            )
        else:
            state = self.restore_checkpoint(
                config.checkpointer.restore_checkpoint_path,
                optimizer=config.optimizer.method,
            )

        loss = Loss(self.predictive, loss_fn=loss_fn)
        loss.rng = self.rng

        if config.monitor.verbose:
            logging.info("Start calibration.")
        state, status = calibrator.train(
            rng=self.rng.get(),
            state=state,
            loss_fun=loss,
            n_epochs=config.optimizer.n_epochs,
            metrics=config.monitor.metrics,
            verbose=config.monitor.verbose,
        )

        self.predictive.state = OutputCalibStateRepository(
            config.checkpointer.save_checkpoint_dir
            if config.checkpointer.dump_state is True
            else None
        )
        self.predictive.state.put(
            state, keep=config.checkpointer.keep_top_n_checkpoints
        )
        return status

    def load_state(self, checkpoint_path: Path) -> None:
        """
        Load a calibration state from a checkpoint path.
        The checkpoint must be compatible with the calibration model.

        Parameters
        ----------
        checkpoint_path : Path
            Path to a checkpoint file or directory to restore.
        """
        try:
            self.restore_checkpoint(checkpoint_path)
        except ValueError:
            raise ValueError(
                f"No checkpoint was found in `checkpoint_path={checkpoint_path}`."
            )
        self.predictive.state = OutputCalibStateRepository(
            checkpoint_dir=checkpoint_path
        )

    def save_state(
        self, checkpoint_path: Path, keep_top_n_checkpoints: int = 1
    ) -> None:
        """
        Save the calibration state as a checkpoint.

        Parameters
        ----------
        checkpoint_path : Path
            Path to file or directory where to save the current state.
        keep_top_n_checkpoints : int
            Number of past checkpoint files to keep.
        """
        if self.predictive.state is None:
            raise ValueError(
                """No state available. You must first either calibrate the model, or load a saved checkpoint."""
            )
        return self.predictive.state.put(
            self.predictive.state.get(),
            checkpoint_path=checkpoint_path,
            keep=keep_top_n_checkpoints,
        )
